# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import str

from ....interfaces import fsl as fsl  # fsl
from ....interfaces import utility as util  # utility
from ....pipeline import engine as pe  # pypeline engine
from ....algorithms import confounds


def select_volume(filename, which):
    """Return the middle index of a file
    """
    from nibabel import load
    import numpy as np
    from nipype.utils import NUMPY_MMAP

    if which.lower() == 'first':
        idx = 0
    elif which.lower() == 'middle':
        idx = int(np.ceil(load(filename, mmap=NUMPY_MMAP).shape[3] / 2))
    else:
        raise Exception('unknown value for volume selection : %s' % which)
    return idx


def create_realign_flow(name='realign'):
    """Realign a time series to the middle volume using spline interpolation

    Uses MCFLIRT to realign the time series and ApplyWarp to apply the rigid
    body transformations using spline interpolation (unknown order).

    Example
    -------

    >>> wf = create_realign_flow()
    >>> wf.inputs.inputspec.func = 'f3.nii'
    >>> wf.run() # doctest: +SKIP

    """
    realignflow = pe.Workflow(name=name)
    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'func',
        ]), name='inputspec')
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'realigned_file',
        ]),
        name='outputspec')
    realigner = pe.Node(
        fsl.MCFLIRT(save_mats=True, stats_imgs=True), name='realigner')
    splitter = pe.Node(fsl.Split(dimension='t'), name='splitter')
    warper = pe.MapNode(
        fsl.ApplyWarp(interp='spline'),
        iterfield=['in_file', 'premat'],
        name='warper')
    joiner = pe.Node(fsl.Merge(dimension='t'), name='joiner')

    realignflow.connect(inputnode, 'func', realigner, 'in_file')
    realignflow.connect(inputnode, ('func', select_volume, 'middle'),
                        realigner, 'ref_vol')
    realignflow.connect(realigner, 'out_file', splitter, 'in_file')
    realignflow.connect(realigner, 'mat_file', warper, 'premat')
    realignflow.connect(realigner, 'variance_img', warper, 'ref_file')
    realignflow.connect(splitter, 'out_files', warper, 'in_file')
    realignflow.connect(warper, 'out_file', joiner, 'in_files')
    realignflow.connect(joiner, 'merged_file', outputnode, 'realigned_file')
    return realignflow


def create_resting_preproc(name='restpreproc', base_dir=None):
    """Create a "resting" time series preprocessing workflow

    The noise removal is based on Behzadi et al. (2007)

    Parameters
    ----------

    name : name of workflow (default: restpreproc)

    Inputs::

        inputspec.func : functional run (filename or list of filenames)

    Outputs::

        outputspec.noise_mask_file : voxels used for PCA to derive noise
                                     components
        outputspec.filtered_file : bandpass filtered and noise-reduced time
                                   series

    Example
    -------

    >>> TR = 3.0
    >>> wf = create_resting_preproc()
    >>> wf.inputs.inputspec.func = 'f3.nii'
    >>> wf.inputs.inputspec.num_noise_components = 6
    >>> wf.inputs.inputspec.highpass_sigma = 100/(2*TR)
    >>> wf.inputs.inputspec.lowpass_sigma = 12.5/(2*TR)
    >>> wf.run() # doctest: +SKIP

    """

    restpreproc = pe.Workflow(name=name, base_dir=base_dir)

    # Define nodes
    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'func', 'num_noise_components', 'highpass_sigma', 'lowpass_sigma'
        ]),
        name='inputspec')
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'noise_mask_file',
            'filtered_file',
        ]),
        name='outputspec')
    slicetimer = pe.Node(fsl.SliceTimer(), name='slicetimer')
    realigner = create_realign_flow()
    tsnr = pe.Node(confounds.TSNR(regress_poly=2), name='tsnr')
    getthresh = pe.Node(
        interface=fsl.ImageStats(op_string='-p 98'), name='getthreshold')
    threshold_stddev = pe.Node(fsl.Threshold(), name='threshold')
    compcor = pe.Node(
        confounds.ACompCor(
            components_file="noise_components.txt", pre_filter=False),
        name='compcor')
    remove_noise = pe.Node(
        fsl.FilterRegressor(filter_all=True), name='remove_noise')
    bandpass_filter = pe.Node(fsl.TemporalFilter(), name='bandpass_filter')

    # Define connections
    restpreproc.connect(inputnode, 'func', slicetimer, 'in_file')
    restpreproc.connect(slicetimer, 'slice_time_corrected_file', realigner,
                        'inputspec.func')
    restpreproc.connect(realigner, 'outputspec.realigned_file', tsnr,
                        'in_file')
    restpreproc.connect(tsnr, 'stddev_file', threshold_stddev, 'in_file')
    restpreproc.connect(tsnr, 'stddev_file', getthresh, 'in_file')
    restpreproc.connect(getthresh, 'out_stat', threshold_stddev, 'thresh')
    restpreproc.connect(realigner, 'outputspec.realigned_file', compcor,
                        'realigned_file')
    restpreproc.connect(threshold_stddev, 'out_file', compcor, 'mask_files')
    restpreproc.connect(inputnode, 'num_noise_components', compcor,
                        'num_components')
    restpreproc.connect(tsnr, 'detrended_file', remove_noise, 'in_file')
    restpreproc.connect(compcor, 'components_file', remove_noise,
                        'design_file')
    restpreproc.connect(inputnode, 'highpass_sigma', bandpass_filter,
                        'highpass_sigma')
    restpreproc.connect(inputnode, 'lowpass_sigma', bandpass_filter,
                        'lowpass_sigma')
    restpreproc.connect(remove_noise, 'out_file', bandpass_filter, 'in_file')
    restpreproc.connect(threshold_stddev, 'out_file', outputnode,
                        'noise_mask_file')
    restpreproc.connect(bandpass_filter, 'out_file', outputnode,
                        'filtered_file')
    return restpreproc



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Example of registration workflows using niftyreg, useful for a variety of
pipelines. Including linear and non-linear image co-registration
"""

from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import str, range
from ....interfaces import utility as niu
from ....interfaces import niftyreg as niftyreg
from ....pipeline import engine as pe


def create_linear_gw_step(name="linear_gw_niftyreg",
                          demean=True,
                          linear_options_hash=None,
                          use_mask=False,
                          verbose=False):
    """
    Creates a workflow that performs linear co-registration of a set of images
    using RegAladin, producing an average image and a set of affine
    transformation matrices linking each of the floating images to the average.

    Inputs::

        inputspec.in_files - The input files to be registered
        inputspec.ref_file - The initial reference image that the input files
                              are registered to
        inputspec.rmask_file - Mask of the reference image
        inputspec.in_aff_files - Initial affine transformation files

    Outputs::

        outputspec.average_image - The average image
        outputspec.aff_files - The affine transformation files

    Optional arguments::

        linear_options_hash - An options dictionary containing a list of
                              parameters for RegAladin that take
        the same form as given in the interface (default None)
        demean - Selects whether to demean the transformation matrices when
                 performing the averaging (default True)
        initial_affines - Selects whether to iterate over initial affine
                          images, which we generally won't have (default False)

    Example
    -------

    >>> from nipype.workflows.smri.niftyreg import create_linear_gw_step
    >>> lgw = create_linear_gw_step('my_linear_coreg')  # doctest: +SKIP
    >>> lgw.inputs.inputspec.in_files = [
    ...     'file1.nii.gz', 'file2.nii.gz']  # doctest: +SKIP
    >>> lgw.inputs.inputspec.ref_file = ['ref.nii.gz']  # doctest: +SKIP
    >>> lgw.run()  # doctest: +SKIP

    """
    # Create the sub workflow
    workflow = pe.Workflow(name=name)
    workflow.base_output_dir = name

    # We need to create an input node for the workflow
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_files', 'ref_file', 'rmask_file']),
        name='inputspec')

    if linear_options_hash is None:
        linear_options_hash = dict()

    # Rigidly register each of the images to the average
    lin_reg = pe.MapNode(
        interface=niftyreg.RegAladin(**linear_options_hash),
        name="lin_reg",
        iterfield=['flo_file'])

    if verbose is False:
        lin_reg.inputs.verbosity_off_flag = True

    # Average the images
    ave_ims = pe.Node(interface=niftyreg.RegAverage(), name="ave_ims")

    # We have a new average image and the affine
    # transformations, which are returned as an output node.
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['average_image', 'trans_files']),
        name='outputspec')

    # Connect the inputs to the lin_reg node
    workflow.connect([(inputnode, lin_reg, [('ref_file', 'ref_file')]),
                      (inputnode, lin_reg, [('in_files', 'flo_file')])])
    if use_mask:
        workflow.connect(inputnode, 'rmask_file', lin_reg, 'rmask_file')

    if demean:
        workflow.connect([(inputnode, ave_ims, [('ref_file',
                                                 'demean1_ref_file')]),
                          (lin_reg, ave_ims, [('avg_output', 'warp_files')])])
    else:
        workflow.connect(lin_reg, 'res_file', ave_ims, 'avg_files')

    # Connect up the output node
    workflow.connect([(lin_reg, outputnode, [('aff_file', 'trans_files')]),
                      (ave_ims, outputnode, [('out_file', 'average_image')])])

    return workflow


def create_nonlinear_gw_step(name="nonlinear_gw_niftyreg",
                             demean=True,
                             nonlinear_options_hash=None,
                             initial_affines=False,
                             use_mask=False,
                             verbose=False):
    """
    Creates a workflow that perform non-linear co-registrations of a set of
    images using RegF3d, producing an non-linear average image and a set of
    cpp transformation linking each of the floating images to the average.

    Inputs::

        inputspec.in_files - The input files to be registered
        inputspec.ref_file - The initial reference image that the input files
                              are registered to
        inputspec.rmask_file - Mask of the reference image
        inputspec.in_trans_files - Initial transformation files (affine or
                                    cpps)

    Outputs::

        outputspec.average_image - The average image
        outputspec.cpp_files - The bspline transformation files

    Optional arguments::

        nonlinear_options_hash - An options dictionary containing a list of
                                 parameters for RegAladin that take the
        same form as given in the interface (default None)
        initial_affines - Selects whether to iterate over initial affine
                          images, which we generally won't have (default False)

    Example
    -------
    >>> from nipype.workflows.smri.niftyreg import create_nonlinear_gw_step
    >>> nlc = create_nonlinear_gw_step('nonlinear_coreg')  # doctest: +SKIP
    >>> nlc.inputs.inputspec.in_files = [
    ...     'file1.nii.gz', 'file2.nii.gz']  # doctest: +SKIP
    >>> nlc.inputs.inputspec.ref_file = ['ref.nii.gz']  # doctest: +SKIP
    >>> nlc.run()  # doctest: +SKIP

    """

    # Create the workflow
    workflow = pe.Workflow(name=name)
    workflow.base_output_dir = name

    # We need to create an input node for the workflow
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['in_files', 'ref_file', 'rmask_file', 'input_aff_files']),
        name='inputspec')

    if nonlinear_options_hash is None:
        nonlinear_options_hash = dict()

    # non-rigidly register each of the images to the average
    # flo_file can take a list of files
    # Need to be able to iterate over input affine files, but what about the
    # cases where we have no input affine files?
    # Passing empty strings are not valid filenames, and undefined fields can
    # not be iterated over.
    # Current simple solution, as this is not generally required, is to use a
    # flag which specifies wherther to iterate
    if initial_affines:
        nonlin_reg = pe.MapNode(
            interface=niftyreg.RegF3D(**nonlinear_options_hash),
            name="nonlin_reg",
            iterfield=['flo_file', 'aff_file'])
    else:
        nonlin_reg = pe.MapNode(
            interface=niftyreg.RegF3D(**nonlinear_options_hash),
            name="nonlin_reg",
            iterfield=['flo_file'])

    if verbose is False:
        nonlin_reg.inputs.verbosity_off_flag = True

    # Average the images
    ave_ims = pe.Node(interface=niftyreg.RegAverage(), name="ave_ims")

    # We have a new centered average image, the resampled original images and
    # the affine transformations, which are returned as an output node.
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['average_image', 'trans_files']),
        name='outputspec')

    # Connect the inputs to the lin_reg node, which is split over in_files
    workflow.connect([(inputnode, nonlin_reg, [('in_files', 'flo_file')]),
                      (inputnode, nonlin_reg, [('ref_file', 'ref_file')])])

    if use_mask:
        workflow.connect(inputnode, 'rmask_file', nonlin_reg, 'rmask_file')

    # If we have initial affine transforms, we need to connect them in
    if initial_affines:
        workflow.connect(inputnode, 'input_aff_files', nonlin_reg, 'aff_file')

    if demean:
        if 'vel_flag' in list(nonlinear_options_hash.keys()) and \
           nonlinear_options_hash['vel_flag'] is True and \
           initial_affines:
            workflow.connect(inputnode, 'ref_file', ave_ims,
                             'demean3_ref_file')
        else:
            workflow.connect(inputnode, 'ref_file', ave_ims,
                             'demean2_ref_file')
        workflow.connect(nonlin_reg, 'avg_output', ave_ims, 'warp_files')
    else:
        workflow.connect(nonlin_reg, 'res_file', ave_ims, 'avg_files')

    # Connect up the output node
    workflow.connect([(nonlin_reg, outputnode, [('cpp_file', 'trans_files')]),
                      (ave_ims, outputnode, [('out_file', 'average_image')])])

    return workflow


# Creates an atlas image by iterative registration. An initial reference image
# can be provided, otherwise one will be made.
def create_groupwise_average(name="atlas_creation",
                             itr_rigid=3,
                             itr_affine=3,
                             itr_non_lin=5,
                             linear_options_hash=None,
                             nonlinear_options_hash=None,
                             use_mask=False,
                             verbose=False):
    """
    Create the overall workflow that embeds all the rigid, affine and
    non-linear components.

    Inputs::

        inputspec.in_files - The input files to be registered
        inputspec.ref_file - The initial reference image that the input files
                              are registered to
        inputspec.rmask_file - Mask of the reference image
        inputspec.in_trans_files - Initial transformation files (affine or
                                    cpps)

    Outputs::

        outputspec.average_image - The average image
        outputspec.cpp_files - The bspline transformation files


    Example
    -------

    >>> from nipype.workflows.smri.niftyreg import create_groupwise_average
    >>> node = create_groupwise_average('groupwise_av')  # doctest: +SKIP
    >>> node.inputs.inputspec.in_files = [
    ...     'file1.nii.gz', 'file2.nii.gz']  # doctest: +SKIP
    >>> node.inputs.inputspec.ref_file = ['ref.nii.gz']  # doctest: +SKIP
    >>> node.inputs.inputspec.rmask_file = ['mask.nii.gz']  # doctest: +SKIP
    >>> node.run()  # doctest: +SKIP

    """
    # Create workflow
    workflow = pe.Workflow(name=name)

    if linear_options_hash is None:
        linear_options_hash = dict()

    if nonlinear_options_hash is None:
        nonlinear_options_hash = dict()

    # Create the input and output node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_files', 'ref_file', 'rmask_file']),
        name='inputspec')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['average_image', 'trans_files']),
        name='outputspec')

    # Create lists to store the rigid, affine and non-linear sub-workflow
    lin_workflows = []
    nonlin_workflows = []

    # Create the linear groupwise registration sub-workflows
    for i in range(itr_rigid + itr_affine):
        # Define is the sub-workflow is rigid or affine
        if i >= itr_rigid:
            linear_options_hash['rig_only_flag'] = False
        else:
            linear_options_hash['rig_only_flag'] = True

        # Define if the average image should be demean to ensure we have a
        # barycenter
        if (i < itr_rigid) or (i == (itr_rigid + itr_affine - 1)):
            demean_arg = False
        else:
            demean_arg = True

        # Create the rigid or affine sub-workflow and add it to the relevant
        # list
        wf = create_linear_gw_step(
            name='lin_reg' + str(i),
            linear_options_hash=linear_options_hash,
            demean=demean_arg,
            verbose=verbose)
        lin_workflows.append(wf)

        # Connect up the input data to the workflow
        workflow.connect(inputnode, 'in_files', wf, 'inputspec.in_files')
        if use_mask:
            workflow.connect(inputnode, 'rmask_file', wf,
                             'inputspec.rmask_file')
        # If it exist, connect the previous workflow to the current one
        if i == 0:
            workflow.connect(inputnode, 'ref_file', wf, 'inputspec.ref_file')
        else:
            workflow.connect(lin_workflows[i - 1], 'outputspec.average_image',
                             wf, 'inputspec.ref_file')

    demean_arg = True

    # Create the nonlinear groupwise registration sub-workflows
    for i in range(itr_non_lin):

        if len(lin_workflows) > 0:
            initial_affines_arg = True
        if i == (itr_non_lin - 1):
            demean_arg = False

        wf = create_nonlinear_gw_step(
            name='nonlin' + str(i),
            demean=demean_arg,
            initial_affines=initial_affines_arg,
            nonlinear_options_hash=nonlinear_options_hash,
            verbose=verbose)

        # Connect up the input data to the workflows
        workflow.connect(inputnode, 'in_files', wf, 'inputspec.in_files')
        if use_mask:
            workflow.connect(inputnode, 'rmask_file', wf,
                             'inputspec.rmask_file')

        if initial_affines_arg:
            # Take the final linear registration results and use them to
            # initialise the NR
            workflow.connect(lin_workflows[-1], 'outputspec.trans_files', wf,
                             'inputspec.input_aff_files')

        if i == 0:
            if len(lin_workflows) > 0:
                workflow.connect(lin_workflows[-1], 'outputspec.average_image',
                                 wf, 'inputspec.ref_file')
            else:
                workflow.connect(inputnode, 'ref_file', wf,
                                 'inputspec.ref_file')
        else:
            workflow.connect(nonlin_workflows[i - 1],
                             'outputspec.average_image', wf,
                             'inputspec.ref_file')

        nonlin_workflows.append(wf)

    # Set up the last workflow
    lw = None
    if len(nonlin_workflows) > 0:
        lw = nonlin_workflows[-1]
    elif len(lin_workflows) > 0:
        lw = lin_workflows[-1]

    # Connect the data to return
    workflow.connect(
        [(lw, outputnode, [('outputspec.average_image', 'average_image')]),
         (lw, outputnode, [('outputspec.trans_files', 'trans_files')])])

    return workflow



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..surface import MergeModels


def test_MergeModels_inputs():
    input_map = dict(
        Model1=dict(
            argstr='%s',
            position=-3,
        ),
        Model2=dict(
            argstr='%s',
            position=-2,
        ),
        ModelOutput=dict(
            argstr='%s',
            hash_files=False,
            position=-1,
        ),
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = MergeModels.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_MergeModels_outputs():
    output_map = dict(ModelOutput=dict(position=-1, ), )
    outputs = MergeModels.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..epi import EpiReg


def test_EpiReg_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        echospacing=dict(argstr='--echospacing=%f', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        epi=dict(
            argstr='--epi=%s',
            mandatory=True,
            position=-4,
        ),
        fmap=dict(argstr='--fmap=%s', ),
        fmapmag=dict(argstr='--fmapmag=%s', ),
        fmapmagbrain=dict(argstr='--fmapmagbrain=%s', ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        no_clean=dict(
            argstr='--noclean',
            usedefault=True,
        ),
        no_fmapreg=dict(argstr='--nofmapreg', ),
        out_base=dict(
            argstr='--out=%s',
            position=-1,
            usedefault=True,
        ),
        output_type=dict(),
        pedir=dict(argstr='--pedir=%s', ),
        t1_brain=dict(
            argstr='--t1brain=%s',
            mandatory=True,
            position=-2,
        ),
        t1_head=dict(
            argstr='--t1=%s',
            mandatory=True,
            position=-3,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
        weight_image=dict(argstr='--weight=%s', ),
        wmseg=dict(argstr='--wmseg=%s', ),
    )
    inputs = EpiReg.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_EpiReg_outputs():
    output_map = dict(
        epi2str_inv=dict(),
        epi2str_mat=dict(),
        fmap2epi_mat=dict(),
        fmap2str_mat=dict(),
        fmap_epi=dict(),
        fmap_str=dict(),
        fmapmag_str=dict(),
        fullwarp=dict(),
        out_1vol=dict(),
        out_file=dict(),
        shiftmap=dict(),
        wmedge=dict(),
        wmseg=dict(),
    )
    outputs = EpiReg.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import MakeAverageSubject


def test_MakeAverageSubject_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        out_name=dict(
            argstr='--out %s',
            usedefault=True,
        ),
        subjects_dir=dict(),
        subjects_ids=dict(
            argstr='--subjects %s',
            mandatory=True,
            sep=' ',
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = MakeAverageSubject.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_MakeAverageSubject_outputs():
    output_map = dict(average_subject_name=dict(), )
    outputs = MakeAverageSubject.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import open

from future.utils import raise_from

import os.path as op

from ....interfaces import io as nio  # Data i/o
from ....interfaces import utility as util  # utility
from ....interfaces import cmtk as cmtk
from ....algorithms import misc as misc
from ....pipeline import engine as pe  # pypeline engine
from ....interfaces.utility import Function
from ....utils.misc import package_check

have_cmp = True
try:
    package_check('cmp')
except Exception as e:
    have_cmp = False
else:
    import cmp


def pullnodeIDs(in_network, name_key='dn_name'):
    """ This function will return the values contained, for each node in
    a network, given an input key. By default it will return the node names
    """
    import networkx as nx
    import numpy as np
    from nipype.interfaces.base import isdefined
    if not isdefined(in_network):
        raise ValueError
        return None
    try:
        ntwk = nx.read_graphml(in_network)
    except:
        ntwk = nx.read_gpickle(in_network)
    nodedata = ntwk.node
    ids = []
    integer_nodelist = []
    for node in list(nodedata.keys()):
        integer_nodelist.append(int(node))
    for node in np.sort(integer_nodelist):
        try:
            nodeid = nodedata[node][name_key]
        except KeyError:
            nodeid = nodedata[str(node)][name_key]
        ids.append(nodeid)
    return ids


def concatcsv(in_files):
    """ This function will contatenate two "comma-separated value"
    text files, but remove the first row (usually column headers) from
    all but the first file.
    """
    import os.path as op
    from nipype.utils.filemanip import split_filename

    if not isinstance(in_files, list):
        return in_files
    if isinstance(in_files[0], list):
        in_files = in_files[0]
    first = open(in_files[0], 'r')
    path, name, ext = split_filename(in_files[0])
    out_name = op.abspath('concat.csv')
    out_file = open(out_name, 'w')
    out_file.write(first.readline())
    first.close()
    for in_file in in_files:
        file_to_read = open(in_file, 'r')
        file_to_read.readline()  # scrap first line
        for line in file_to_read:
            out_file.write(line)
    return out_name


def create_merge_networks_by_group_workflow(group_list, group_id, data_dir,
                                            subjects_dir, output_dir):
    """Creates a second-level pipeline to merge the Connectome File Format (CFF) outputs from the group-level
    MRtrix structural connectivity processing pipeline into a single CFF file for each group.

    Example
    -------

    >>> import nipype.workflows.dmri.connectivity.group_connectivity as groupwork
    >>> from nipype.testing import example_data
    >>> subjects_dir = '.'
    >>> data_dir = '.'
    >>> output_dir = '.'
    >>> group_list = {}
    >>> group_list['group1'] = ['subj1', 'subj2']
    >>> group_list['group2'] = ['subj3', 'subj4']
    >>> group_id = 'group1'
    >>> l2pipeline = groupwork.create_merge_networks_by_group_workflow(group_list, group_id, data_dir, subjects_dir, output_dir)
    >>> l2pipeline.run()                 # doctest: +SKIP

    Inputs::

        group_list: Dictionary of subject lists, keyed by group name
        group_id: String containing the group name
        data_dir: Path to the data directory
        subjects_dir: Path to the Freesurfer 'subjects' directory
        output_dir: Path for the output files
    """
    group_infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id']),
        name="group_infosource")
    group_infosource.inputs.group_id = group_id

    l2infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id']),
        name='l2infosource')

    l2source = pe.Node(
        nio.DataGrabber(infields=['group_id'], outfields=['CFFfiles']),
        name='l2source')
    l2source.inputs.template_args = dict(CFFfiles=[['group_id']])
    l2source.inputs.template = op.join(output_dir, '%s/cff/*/connectome.cff')
    l2source.inputs.base_directory = data_dir
    l2source.inputs.sort_filelist = True

    l2inputnode = pe.Node(
        interface=util.IdentityInterface(fields=['CFFfiles']),
        name='l2inputnode')
    MergeCNetworks = pe.Node(
        interface=cmtk.MergeCNetworks(), name="MergeCNetworks")

    l2datasink = pe.Node(interface=nio.DataSink(), name="l2datasink")
    l2datasink.inputs.base_directory = output_dir
    l2datasink.inputs.container = group_id

    l2pipeline = pe.Workflow(name="l2output_" + group_id)
    l2pipeline.base_dir = op.join(output_dir, 'l2output')
    l2pipeline.connect([(group_infosource, l2infosource, [('group_id',
                                                           'group_id')])])

    l2pipeline.connect([
        (l2infosource, l2source, [('group_id', 'group_id')]),
        (l2source, l2inputnode, [('CFFfiles', 'CFFfiles')]),
    ])

    l2pipeline.connect([(l2inputnode, MergeCNetworks, [('CFFfiles',
                                                        'in_files')])])
    l2pipeline.connect([(group_infosource, MergeCNetworks, [('group_id',
                                                             'out_file')])])
    l2pipeline.connect([(MergeCNetworks, l2datasink, [('connectome_file',
                                                       '@l2output')])])
    l2pipeline.connect([(group_infosource, l2datasink, [('group_id',
                                                         '@group_id')])])
    return l2pipeline


def create_merge_network_results_by_group_workflow(
        group_list, group_id, data_dir, subjects_dir, output_dir):
    """Creates a second-level pipeline to merge the Connectome File Format (CFF) outputs from the group-level
    MRtrix structural connectivity processing pipeline into a single CFF file for each group.

    Example
    -------

    >>> import nipype.workflows.dmri.connectivity.group_connectivity as groupwork
    >>> from nipype.testing import example_data
    >>> subjects_dir = '.'
    >>> data_dir = '.'
    >>> output_dir = '.'
    >>> group_list = {}
    >>> group_list['group1'] = ['subj1', 'subj2']
    >>> group_list['group2'] = ['subj3', 'subj4']
    >>> group_id = 'group1'
    >>> l2pipeline = groupwork.create_merge_network_results_by_group_workflow(group_list, group_id, data_dir, subjects_dir, output_dir)
    >>> l2pipeline.run()                 # doctest: +SKIP

    Inputs::

        group_list: Dictionary of subject lists, keyed by group name
        group_id: String containing the group name
        data_dir: Path to the data directory
        subjects_dir: Path to the Freesurfer 'subjects' directory
        output_dir: Path for the output files
    """
    group_infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id']),
        name="group_infosource")
    group_infosource.inputs.group_id = group_id

    l2infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id', 'merged']),
        name='l2infosource')

    l2source = pe.Node(
        nio.DataGrabber(
            infields=['group_id'],
            outfields=[
                'CFFfiles', 'CSVmatrices', 'CSVfibers', 'CSVnodal', 'CSVglobal'
            ]),
        name='l2source')

    l2source.inputs.template_args = dict(
        CFFfiles=[['group_id']],
        CSVmatrices=[['group_id']],
        CSVnodal=[['group_id']],
        CSVglobal=[['group_id']],
        CSVfibers=[['group_id']])
    l2source.inputs.base_directory = data_dir
    l2source.inputs.template = '%s/%s'
    l2source.inputs.field_template = dict(
        CFFfiles=op.join(output_dir, '%s/cff/*/connectome.cff'),
        CSVmatrices=op.join(output_dir, '%s/cmatrices_csv/*/*.csv'),
        CSVnodal=op.join(output_dir, '%s/nxcsv/*/*nodal*.csv'),
        CSVglobal=op.join(output_dir, '%s/nxcsv/*/*global*.csv'),
        CSVfibers=op.join(output_dir, '%s/fiber_csv/*/*fibers*.csv'))
    l2source.inputs.sort_filelist = True

    l2inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'CFFfiles', 'CSVfibers', 'CSVmatrices', 'CSVnodal', 'CSVglobal',
            'network_file'
        ]),
        name='l2inputnode')

    MergeCNetworks = pe.Node(
        interface=cmtk.MergeCNetworks(), name="MergeCNetworks")

    l2datasink = pe.Node(interface=nio.DataSink(), name="l2datasink")
    l2datasink.inputs.base_directory = output_dir
    l2datasink.inputs.container = group_id

    l2pipeline = pe.Workflow(name="l2output_" + group_id)
    l2pipeline.base_dir = op.join(output_dir, 'l2output')
    l2pipeline.connect([(group_infosource, l2infosource, [('group_id',
                                                           'group_id')])])

    l2pipeline.connect([
        (l2infosource, l2source, [('group_id', 'group_id')]),
        (l2source, l2inputnode, [('CFFfiles', 'CFFfiles')]),
        (l2source, l2inputnode, [('CSVmatrices', 'CSVmatrices')]),
        (l2source, l2inputnode, [('CSVnodal', 'CSVnodal')]),
        (l2source, l2inputnode, [('CSVglobal', 'CSVglobal')]),
        (l2source, l2inputnode, [('CSVfibers', 'CSVfibers')]),
    ])

    l2pipeline.connect([(l2inputnode, MergeCNetworks, [('CFFfiles',
                                                        'in_files')])])

    l2pipeline.connect([(group_infosource, MergeCNetworks, [('group_id',
                                                             'out_file')])])
    l2pipeline.connect([(MergeCNetworks, l2datasink, [('connectome_file',
                                                       '@l2output')])])

    AddCSVColumn_node = pe.Node(
        interface=misc.AddCSVColumn(), name="AddCSVColumn_node")
    AddCSVColumn_node.inputs.extra_column_heading = 'group'
    AddCSVColumn_global = AddCSVColumn_node.clone(name="AddCSVColumn_global")
    AddCSVColumn_matrices = AddCSVColumn_node.clone(
        name="AddCSVColumn_matrices")
    AddCSVColumn_fibers = AddCSVColumn_node.clone(name="AddCSVColumn_fibers")

    concat_csv_interface = Function(
        input_names=["in_files"],
        output_names=["out_name"],
        function=concatcsv)

    concat_node_csvs = pe.Node(
        interface=concat_csv_interface, name='concat_node_csvs')
    concat_global_csvs = pe.Node(
        interface=concat_csv_interface, name='concat_global_csvs')
    concat_matrix_csvs = pe.Node(
        interface=concat_csv_interface, name='concat_matrix_csvs')
    concat_fiber_csvs = pe.Node(
        interface=concat_csv_interface, name='concat_fiber_csvs')

    l2pipeline.connect([(l2inputnode, concat_node_csvs, [('CSVnodal',
                                                          'in_files')])])
    l2pipeline.connect([(concat_node_csvs, AddCSVColumn_node, [('out_name',
                                                                'in_file')])])
    l2pipeline.connect([(group_infosource, AddCSVColumn_node,
                         [('group_id', 'extra_field')])])
    l2pipeline.connect([(AddCSVColumn_node, l2datasink,
                         [('csv_file', '@l2output.node_csv')])])
    l2pipeline.connect([(group_infosource, l2datasink, [('group_id',
                                                         '@group_id')])])

    l2pipeline.connect([(l2inputnode, concat_global_csvs, [('CSVglobal',
                                                            'in_files')])])
    l2pipeline.connect([(concat_global_csvs, AddCSVColumn_global,
                         [('out_name', 'in_file')])])
    l2pipeline.connect([(group_infosource, AddCSVColumn_global,
                         [('group_id', 'extra_field')])])
    l2pipeline.connect([(AddCSVColumn_global, l2datasink,
                         [('csv_file', '@l2output.global_csv')])])

    l2pipeline.connect([(l2inputnode, concat_matrix_csvs, [('CSVmatrices',
                                                            'in_files')])])
    l2pipeline.connect([(concat_matrix_csvs, AddCSVColumn_matrices,
                         [('out_name', 'in_file')])])
    l2pipeline.connect([(group_infosource, AddCSVColumn_matrices,
                         [('group_id', 'extra_field')])])
    l2pipeline.connect([(AddCSVColumn_matrices, l2datasink,
                         [('csv_file', '@l2output.cmatrices_csv')])])

    l2pipeline.connect([(l2inputnode, concat_fiber_csvs, [('CSVmatrices',
                                                           'in_files')])])
    l2pipeline.connect([(concat_fiber_csvs, AddCSVColumn_fibers,
                         [('out_name', 'in_file')])])
    l2pipeline.connect([(group_infosource, AddCSVColumn_fibers,
                         [('group_id', 'extra_field')])])
    l2pipeline.connect([(AddCSVColumn_fibers, l2datasink,
                         [('csv_file', '@l2output.fibers_csv')])])
    return l2pipeline


def create_merge_group_networks_workflow(group_list,
                                         data_dir,
                                         subjects_dir,
                                         output_dir,
                                         title='group'):
    """Creates a third-level pipeline to merge the Connectome File Format (CFF) outputs from each group
    and combines them into a single CFF file for each group.

    Example
    -------

    >>> import nipype.workflows.dmri.connectivity.group_connectivity as groupwork
    >>> from nipype.testing import example_data
    >>> subjects_dir = '.'
    >>> data_dir = '.'
    >>> output_dir = '.'
    >>> group_list = {}
    >>> group_list['group1'] = ['subj1', 'subj2']
    >>> group_list['group2'] = ['subj3', 'subj4']
    >>> l3pipeline = groupwork.create_merge_group_networks_workflow(group_list, data_dir, subjects_dir, output_dir)
    >>> l3pipeline.run()                 # doctest: +SKIP

    Inputs::

        group_list: Dictionary of subject lists, keyed by group name
        data_dir: Path to the data directory
        subjects_dir: Path to the Freesurfer 'subjects' directory
        output_dir: Path for the output files
        title: String to use as a title for the output merged CFF file (default 'group')
    """
    l3infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id']),
        name='l3infosource')
    l3infosource.inputs.group_id = list(group_list.keys())

    l3source = pe.Node(
        nio.DataGrabber(infields=['group_id'], outfields=['CFFfiles']),
        name='l3source')
    l3source.inputs.template_args = dict(CFFfiles=[['group_id', 'group_id']])
    l3source.inputs.template = op.join(output_dir, '%s/%s.cff')
    l3source.inputs.sort_filelist = True

    l3inputnode = pe.Node(
        interface=util.IdentityInterface(fields=['Group_CFFs']),
        name='l3inputnode')

    MergeCNetworks_grp = pe.Node(
        interface=cmtk.MergeCNetworks(), name="MergeCNetworks_grp")
    MergeCNetworks_grp.inputs.out_file = title

    l3datasink = pe.Node(interface=nio.DataSink(), name="l3datasink")
    l3datasink.inputs.base_directory = output_dir

    l3pipeline = pe.Workflow(name="l3output")
    l3pipeline.base_dir = output_dir
    l3pipeline.connect([
        (l3infosource, l3source, [('group_id', 'group_id')]),
        (l3source, l3inputnode, [('CFFfiles', 'Group_CFFs')]),
    ])

    l3pipeline.connect([(l3inputnode, MergeCNetworks_grp, [('Group_CFFs',
                                                            'in_files')])])
    l3pipeline.connect([(MergeCNetworks_grp, l3datasink, [('connectome_file',
                                                           '@l3output')])])
    return l3pipeline


def create_merge_group_network_results_workflow(group_list,
                                                data_dir,
                                                subjects_dir,
                                                output_dir,
                                                title='group'):
    """Creates a third-level pipeline to merge the Connectome File Format (CFF) outputs from each group
    and combines them into a single CFF file for each group. This version of the third-level pipeline also
    concatenates the comma-separated value files for the NetworkX metrics and the connectivity matrices
    into single files.

    Example
    -------

    >>> import nipype.workflows.dmri.connectivity.group_connectivity as groupwork
    >>> from nipype.testing import example_data
    >>> subjects_dir = '.'
    >>> data_dir = '.'
    >>> output_dir = '.'
    >>> group_list = {}
    >>> group_list['group1'] = ['subj1', 'subj2']
    >>> group_list['group2'] = ['subj3', 'subj4']
    >>> l3pipeline = groupwork.create_merge_group_network_results_workflow(group_list, data_dir, subjects_dir, output_dir)
    >>> l3pipeline.run()                 # doctest: +SKIP

    Inputs::

        group_list: Dictionary of subject lists, keyed by group name
        data_dir: Path to the data directory
        subjects_dir: Path to the Freesurfer 'subjects' directory
        output_dir: Path for the output files
        title: String to use as a title for the output merged CFF file (default 'group')
    """
    l3infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id']),
        name='l3infosource')
    l3infosource.inputs.group_id = list(group_list.keys())

    l3source = pe.Node(
        nio.DataGrabber(
            infields=['group_id'],
            outfields=[
                'CFFfiles', 'CSVnodemetrics', 'CSVglobalmetrics', 'CSVmatrices'
            ]),
        name='l3source')
    l3source.inputs.template_args = dict(
        CFFfiles=[['group_id']],
        CSVnodemetrics=[['group_id']],
        CSVglobalmetrics=[['group_id']],
        CSVmatrices=[['group_id']])
    l3source.inputs.template = op.join(output_dir, '%s/%s')
    l3source.inputs.sort_filelist = True

    l3source.inputs.field_template = dict(
        CFFfiles=op.join(output_dir, '%s/*.cff'),
        CSVnodemetrics=op.join(output_dir, '%s/node_csv/*.csv'),
        CSVglobalmetrics=op.join(output_dir, '%s/global_csv/*.csv'),
        CSVmatrices=op.join(output_dir, '%s/cmatrices_csv/*/*.csv'))

    l3inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'Group_CFFs', 'Group_CSVnodemetrics', 'Group_CSVglobalmetrics',
            'Group_CSVmatrices'
        ]),
        name='l3inputnode')

    MergeCNetworks_grp = pe.Node(
        interface=cmtk.MergeCNetworks(), name="MergeCNetworks_grp")
    MergeCNetworks_grp.inputs.out_file = title

    l3datasink = pe.Node(interface=nio.DataSink(), name="l3datasink")
    l3datasink.inputs.base_directory = output_dir

    l3pipeline = pe.Workflow(name="l3output")
    l3pipeline.base_dir = output_dir
    l3pipeline.connect([
        (l3infosource, l3source, [('group_id', 'group_id')]),
        (l3source, l3inputnode, [('CFFfiles', 'Group_CFFs')]),
        (l3source, l3inputnode, [('CSVnodemetrics', 'Group_CSVnodemetrics')]),
        (l3source, l3inputnode, [('CSVglobalmetrics',
                                  'Group_CSVglobalmetrics')]),
        (l3source, l3inputnode, [('CSVmatrices', 'Group_CSVmatrices')]),
    ])

    l3pipeline.connect([(l3inputnode, MergeCNetworks_grp, [('Group_CFFs',
                                                            'in_files')])])
    l3pipeline.connect([(MergeCNetworks_grp, l3datasink, [('connectome_file',
                                                           '@l3output')])])

    concat_csv_interface = Function(
        input_names=["in_files"],
        output_names=["out_name"],
        function=concatcsv)

    concat_node_csvs = pe.Node(
        interface=concat_csv_interface, name='concat_node_csvs')
    concat_global_csvs = pe.Node(
        interface=concat_csv_interface, name='concat_global_csvs')
    concat_matrix_csvs = pe.Node(
        interface=concat_csv_interface, name='concat_matrix_csvs')

    l3pipeline.connect([(l3inputnode, concat_node_csvs,
                         [('Group_CSVnodemetrics', 'in_files')])])
    l3pipeline.connect([(concat_node_csvs, l3datasink,
                         [('out_name', '@l3output.nodal_csv')])])

    l3pipeline.connect([(l3inputnode, concat_global_csvs,
                         [('Group_CSVglobalmetrics', 'in_files')])])
    l3pipeline.connect([(concat_global_csvs, l3datasink,
                         [('out_name', '@l3output.global_csv')])])

    l3pipeline.connect([(l3inputnode, concat_matrix_csvs,
                         [('Group_CSVmatrices', 'in_files')])])
    l3pipeline.connect([(concat_matrix_csvs, l3datasink,
                         [('out_name', '@l3output.csvmatrices')])])
    return l3pipeline


def create_average_networks_by_group_workflow(group_list,
                                              data_dir,
                                              subjects_dir,
                                              output_dir,
                                              title='group_average'):
    """Creates a fourth-level pipeline to average the networks for two groups and merge them into a single
    CFF file. This pipeline will also output the average networks in .gexf format, for visualization in other
    graph viewers, such as Gephi.

    Example
    -------

    >>> import nipype.workflows.dmri.connectivity.group_connectivity as groupwork
    >>> from nipype.testing import example_data
    >>> subjects_dir = '.'
    >>> data_dir = '.'
    >>> output_dir = '.'
    >>> group_list = {}
    >>> group_list['group1'] = ['subj1', 'subj2']
    >>> group_list['group2'] = ['subj3', 'subj4']
    >>> l4pipeline = groupwork.create_average_networks_by_group_workflow(group_list, data_dir, subjects_dir, output_dir)
    >>> l4pipeline.run()                 # doctest: +SKIP

    Inputs::

        group_list: Dictionary of subject lists, keyed by group name
        data_dir: Path to the data directory
        subjects_dir: Path to the Freesurfer 'subjects' directory
        output_dir: Path for the output files
        title: String to use as a title for the output merged CFF file (default 'group')
    """
    l4infosource = pe.Node(
        interface=util.IdentityInterface(fields=['group_id1', 'group_id2']),
        name='l4infosource')
    try:
        l4infosource.inputs.group_id1 = list(group_list.keys())[0]
        l4infosource.inputs.group_id2 = list(group_list.keys())[1]
    except IndexError as e:
        raise_from(
            Exception(
                'The create_average_networks_by_group_workflow requires 2 groups'
            ), e)

    l4info = dict(
        networks=[['group_id', '']],
        CMatrices=[['group_id', '']],
        fibmean=[['group_id', 'mean_fiber_length']],
        fibdev=[['group_id', 'fiber_length_std']])

    l4source_grp1 = pe.Node(
        nio.DataGrabber(infields=['group_id'], outfields=list(l4info.keys())),
        name='l4source_grp1')
    l4source_grp1.inputs.template = '%s/%s'
    l4source_grp1.inputs.field_template = dict(
        networks=op.join(output_dir, '%s/networks/*/*%s*intersections*.pck'),
        CMatrices=op.join(output_dir, '%s/cmatrix/*/*%s*.mat'),
        fibmean=op.join(output_dir, '%s/mean_fiber_length/*/*%s*.mat'),
        fibdev=op.join(output_dir, '%s/fiber_length_std/*/*%s*.mat'))
    l4source_grp1.inputs.base_directory = output_dir
    l4source_grp1.inputs.template_args = l4info
    l4source_grp1.inputs.sort_filelist = True

    l4source_grp2 = l4source_grp1.clone(name='l4source_grp2')

    l4inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'networks_grp1', 'networks_grp2', 'CMatrices_grp1',
            'CMatrices_grp2', 'fibmean_grp1', 'fibmean_grp2', 'fibdev_grp1',
            'fibdev_grp2'
        ]),
        name='l4inputnode')

    average_networks_grp1 = pe.Node(
        interface=cmtk.AverageNetworks(), name='average_networks_grp1')
    average_networks_grp2 = average_networks_grp1.clone(
        'average_networks_grp2')

    averagecff = pe.Node(interface=cmtk.CFFConverter(), name="averagecff")
    averagecff.inputs.out_file = title

    merge_gpickled_averages = pe.Node(
        interface=util.Merge(2), name='merge_gpickled_averages')
    merge_gexf_averages = merge_gpickled_averages.clone('merge_gexf_averages')

    l4datasink = pe.Node(interface=nio.DataSink(), name="l4datasink")
    l4datasink.inputs.base_directory = output_dir

    l4pipeline = pe.Workflow(name="l4output")
    l4pipeline.base_dir = output_dir
    l4pipeline.connect([
        (l4infosource, l4source_grp1, [('group_id1', 'group_id')]),
        (l4infosource, l4source_grp2, [('group_id2', 'group_id')]),
        (l4source_grp1, l4inputnode, [('CMatrices', 'CMatrices_grp1')]),
        (l4source_grp2, l4inputnode, [('CMatrices', 'CMatrices_grp2')]),
        (l4source_grp1, l4inputnode, [('networks', 'networks_grp1')]),
        (l4source_grp2, l4inputnode, [('networks', 'networks_grp2')]),
        (l4source_grp1, l4inputnode, [('fibmean', 'fibmean_grp1')]),
        (l4source_grp2, l4inputnode, [('fibmean', 'fibmean_grp2')]),
        (l4source_grp1, l4inputnode, [('fibdev', 'fibdev_grp1')]),
        (l4source_grp2, l4inputnode, [('fibdev', 'fibdev_grp2')]),
    ])

    l4pipeline.connect([(l4inputnode, average_networks_grp1, [('networks_grp1',
                                                               'in_files')])])
    l4pipeline.connect([(l4infosource, average_networks_grp1, [('group_id1',
                                                                'group_id')])])

    l4pipeline.connect([(l4inputnode, average_networks_grp2, [('networks_grp2',
                                                               'in_files')])])
    l4pipeline.connect([(l4infosource, average_networks_grp2, [('group_id2',
                                                                'group_id')])])

    l4pipeline.connect([(average_networks_grp1, merge_gpickled_averages,
                         [('gpickled_groupavg', 'in1')])])
    l4pipeline.connect([(average_networks_grp2, merge_gpickled_averages,
                         [('gpickled_groupavg', 'in2')])])

    l4pipeline.connect([(average_networks_grp1, merge_gexf_averages,
                         [('gexf_groupavg', 'in1')])])
    l4pipeline.connect([(average_networks_grp2, merge_gexf_averages,
                         [('gexf_groupavg', 'in2')])])

    l4pipeline.connect([(merge_gpickled_averages, l4datasink,
                         [('out', '@l4output.gpickled')])])
    l4pipeline.connect([(merge_gpickled_averages, averagecff,
                         [('out', 'gpickled_networks')])])
    l4pipeline.connect([(averagecff, l4datasink, [('connectome_file',
                                                   '@l4output.averagecff')])])

    l4pipeline.connect([(merge_gexf_averages, l4datasink,
                         [('out', '@l4output.gexf')])])
    return l4pipeline



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..confounds import TCompCor


def test_TCompCor_inputs():
    input_map = dict(
        components_file=dict(usedefault=True, ),
        header_prefix=dict(),
        high_pass_cutoff=dict(usedefault=True, ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        ignore_initial_volumes=dict(usedefault=True, ),
        mask_files=dict(),
        mask_index=dict(
            requires=['mask_files'],
            xor=['merge_method'],
        ),
        merge_method=dict(
            requires=['mask_files'],
            xor=['mask_index'],
        ),
        num_components=dict(usedefault=True, ),
        percentile_threshold=dict(usedefault=True, ),
        pre_filter=dict(usedefault=True, ),
        realigned_file=dict(mandatory=True, ),
        regress_poly_degree=dict(usedefault=True, ),
        repetition_time=dict(),
        save_pre_filter=dict(),
        use_regress_poly=dict(
            deprecated='0.15.0',
            new_name='pre_filter',
        ),
    )
    inputs = TCompCor.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_TCompCor_outputs():
    output_map = dict(
        components_file=dict(),
        high_variance_masks=dict(),
        pre_filter_file=dict(),
    )
    outputs = TCompCor.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
import inspect
import os.path as op

from ....interfaces import io as nio  # Data i/o
from ....interfaces import utility as util  # utility
from ....pipeline import engine as pe  # pypeline engine
from ....interfaces import camino as camino
from ....interfaces import fsl as fsl
from ....interfaces import camino2trackvis as cam2trk
from ....interfaces import freesurfer as fs  # freesurfer
from ....interfaces import cmtk as cmtk
from ....algorithms import misc as misc
from ...misc.utils import (get_affine, get_data_dims, get_vox_dims,
                           select_aparc, select_aparc_annot)


def create_connectivity_pipeline(name="connectivity"):
    """Creates a pipeline that does the same connectivity processing as in the
    :ref:`example_dmri_connectivity` example script. Given a subject id (and completed Freesurfer reconstruction)
    diffusion-weighted image, b-values, and b-vectors, the workflow will return the subject's connectome
    as a Connectome File Format (CFF) file for use in Connectome Viewer (http://www.cmtk.org).

    Example
    -------

    >>> from nipype.workflows.dmri.camino.connectivity_mapping import create_connectivity_pipeline
    >>> conmapper = create_connectivity_pipeline("nipype_conmap")
    >>> conmapper.inputs.inputnode.subjects_dir = '.'
    >>> conmapper.inputs.inputnode.subject_id = 'subj1'
    >>> conmapper.inputs.inputnode.dwi = 'data.nii.gz'
    >>> conmapper.inputs.inputnode.bvecs = 'bvecs'
    >>> conmapper.inputs.inputnode.bvals = 'bvals'
    >>> conmapper.run()                 # doctest: +SKIP

    Inputs::

        inputnode.subject_id
        inputnode.subjects_dir
        inputnode.dwi
        inputnode.bvecs
        inputnode.bvals
        inputnode.resolution_network_file

    Outputs::

        outputnode.connectome
        outputnode.cmatrix
        outputnode.gpickled_network
        outputnode.fa
        outputnode.struct
        outputnode.trace
        outputnode.tracts
        outputnode.tensors

    """

    inputnode_within = pe.Node(
        interface=util.IdentityInterface(fields=[
            "subject_id",
            "dwi",
            "bvecs",
            "bvals",
            "subjects_dir",
            "resolution_network_file",
        ]),
        name="inputnode_within")

    FreeSurferSource = pe.Node(
        interface=nio.FreeSurferSource(), name='fssource')

    FreeSurferSourceLH = pe.Node(
        interface=nio.FreeSurferSource(), name='fssourceLH')
    FreeSurferSourceLH.inputs.hemi = 'lh'

    FreeSurferSourceRH = pe.Node(
        interface=nio.FreeSurferSource(), name='fssourceRH')
    FreeSurferSourceRH.inputs.hemi = 'rh'
    """
    Since the b values and b vectors come from the FSL course, we must convert it to a scheme file
    for use in Camino.
    """

    fsl2scheme = pe.Node(interface=camino.FSL2Scheme(), name="fsl2scheme")
    fsl2scheme.inputs.usegradmod = True
    """
    FSL's Brain Extraction tool is used to create a mask from the b0 image
    """

    b0Strip = pe.Node(interface=fsl.BET(mask=True), name='bet_b0')
    """
    FSL's FLIRT function is used to coregister the b0 mask and the structural image.
    A convert_xfm node is then used to obtain the inverse of the transformation matrix.
    FLIRT is used once again to apply the inverse transformation to the parcellated brain image.
    """

    coregister = pe.Node(interface=fsl.FLIRT(dof=6), name='coregister')
    coregister.inputs.cost = ('normmi')

    convertxfm = pe.Node(interface=fsl.ConvertXFM(), name='convertxfm')
    convertxfm.inputs.invert_xfm = True

    inverse = pe.Node(interface=fsl.FLIRT(), name='inverse')
    inverse.inputs.interp = ('nearestneighbour')

    inverse_AparcAseg = pe.Node(
        interface=fsl.FLIRT(), name='inverse_AparcAseg')
    inverse_AparcAseg.inputs.interp = ('nearestneighbour')
    """
    A number of conversion operations are required to obtain NIFTI files from the FreesurferSource for each subject.
    Nodes are used to convert the following:
        * Original structural image to NIFTI
        * Parcellated white matter image to NIFTI
        * Parcellated whole-brain image to NIFTI
        * Pial, white, inflated, and spherical surfaces for both the left and right hemispheres
            are converted to GIFTI for visualization in ConnectomeViewer
        * Parcellated annotation files for the left and right hemispheres are also converted to GIFTI
    """

    mri_convert_Brain = pe.Node(
        interface=fs.MRIConvert(), name='mri_convert_Brain')
    mri_convert_Brain.inputs.out_type = 'nii'

    mri_convert_AparcAseg = mri_convert_Brain.clone('mri_convert_AparcAseg')

    mris_convertLH = pe.Node(interface=fs.MRIsConvert(), name='mris_convertLH')
    mris_convertLH.inputs.out_datatype = 'gii'
    mris_convertRH = mris_convertLH.clone('mris_convertRH')
    mris_convertRHwhite = mris_convertLH.clone('mris_convertRHwhite')
    mris_convertLHwhite = mris_convertLH.clone('mris_convertLHwhite')
    mris_convertRHinflated = mris_convertLH.clone('mris_convertRHinflated')
    mris_convertLHinflated = mris_convertLH.clone('mris_convertLHinflated')
    mris_convertRHsphere = mris_convertLH.clone('mris_convertRHsphere')
    mris_convertLHsphere = mris_convertLH.clone('mris_convertLHsphere')
    mris_convertLHlabels = mris_convertLH.clone('mris_convertLHlabels')
    mris_convertRHlabels = mris_convertLH.clone('mris_convertRHlabels')
    """
    In this section we create the nodes necessary for diffusion analysis.
    First, the diffusion image is converted to voxel order, since this is the format in which Camino does
    its processing.
    """

    image2voxel = pe.Node(interface=camino.Image2Voxel(), name="image2voxel")
    """
    Second, diffusion tensors are fit to the voxel-order data.
    If desired, these tensors can be converted to a Nifti tensor image using the DT2NIfTI interface.
    """

    dtifit = pe.Node(interface=camino.DTIFit(), name='dtifit')
    """
    Next, a lookup table is generated from the schemefile and the
    signal-to-noise ratio (SNR) of the unweighted (q=0) data.
    """

    dtlutgen = pe.Node(interface=camino.DTLUTGen(), name="dtlutgen")
    dtlutgen.inputs.snr = 16.0
    dtlutgen.inputs.inversion = 1
    """
    In this tutorial we implement probabilistic tractography using the PICo algorithm.
    PICo tractography requires an estimate of the fibre direction and a model of its uncertainty in each voxel;
    this probabilitiy distribution map is produced using the following node.
    """

    picopdfs = pe.Node(interface=camino.PicoPDFs(), name="picopdfs")
    picopdfs.inputs.inputmodel = 'dt'
    """
    Finally, tractography is performed. In this tutorial, we will use only one iteration for time-saving purposes.
    It is important to note that we use the TrackPICo interface here. This interface now expects the files required
    for PICo tracking (i.e. the output from picopdfs). Similar interfaces exist for alternative types of tracking,
    such as Bayesian tracking with Dirac priors (TrackBayesDirac).
    """

    track = pe.Node(interface=camino.TrackPICo(), name="track")
    track.inputs.iterations = 1
    """
    Currently, the best program for visualizing tracts is TrackVis. For this reason, a node is included to
    convert the raw tract data to .trk format. Solely for testing purposes, another node is added to perform the reverse.
    """

    camino2trackvis = pe.Node(
        interface=cam2trk.Camino2Trackvis(), name="camino2trackvis")
    camino2trackvis.inputs.min_length = 30
    camino2trackvis.inputs.voxel_order = 'LAS'
    trk2camino = pe.Node(
        interface=cam2trk.Trackvis2Camino(), name="trk2camino")
    """
    Tracts can also be converted to VTK and OOGL formats, for use in programs such as GeomView and Paraview,
    using the following two nodes.
    """

    vtkstreamlines = pe.Node(
        interface=camino.VtkStreamlines(), name="vtkstreamlines")
    procstreamlines = pe.Node(
        interface=camino.ProcStreamlines(), name="procstreamlines")
    """
    We can easily produce a variety of scalar values from our fitted tensors. The following nodes generate the
    fractional anisotropy and diffusivity trace maps and their associated headers, and then merge them back
    into a single .nii file.
    """

    fa = pe.Node(interface=camino.ComputeFractionalAnisotropy(), name='fa')
    trace = pe.Node(interface=camino.ComputeTensorTrace(), name='trace')
    dteig = pe.Node(interface=camino.ComputeEigensystem(), name='dteig')

    analyzeheader_fa = pe.Node(
        interface=camino.AnalyzeHeader(), name='analyzeheader_fa')
    analyzeheader_fa.inputs.datatype = 'double'
    analyzeheader_trace = pe.Node(
        interface=camino.AnalyzeHeader(), name='analyzeheader_trace')
    analyzeheader_trace.inputs.datatype = 'double'

    fa2nii = pe.Node(interface=misc.CreateNifti(), name='fa2nii')
    trace2nii = fa2nii.clone("trace2nii")
    """
    This section adds the Connectome Mapping Toolkit (CMTK) nodes.
    These interfaces are fairly experimental and may not function properly.
    In order to perform connectivity mapping using CMTK, the parcellated structural data is rewritten
    using the indices and parcellation scheme from the connectome mapper (CMP). This process has been
    written into the ROIGen interface, which will output a remapped aparc+aseg image as well as a
    dictionary of label information (i.e. name, display colours) pertaining to the original and remapped regions.
    These label values are input from a user-input lookup table, if specified, and otherwise the default
    Freesurfer LUT (/freesurfer/FreeSurferColorLUT.txt).
    """

    roigen = pe.Node(interface=cmtk.ROIGen(), name="ROIGen")
    roigen_structspace = roigen.clone("ROIGen_structspace")
    """
    The CreateMatrix interface takes in the remapped aparc+aseg image as well as the label dictionary and fiber tracts
    and outputs a number of different files. The most important of which is the connectivity network itself, which is stored
    as a 'gpickle' and can be loaded using Python's NetworkX package (see CreateMatrix docstring). Also outputted are various
    NumPy arrays containing detailed tract information, such as the start and endpoint regions, and statistics on the mean and
    standard deviation for the fiber length of each connection. These matrices can be used in the ConnectomeViewer to plot the
    specific tracts that connect between user-selected regions.
    """

    createnodes = pe.Node(interface=cmtk.CreateNodes(), name="CreateNodes")
    creatematrix = pe.Node(interface=cmtk.CreateMatrix(), name="CreateMatrix")
    creatematrix.inputs.count_region_intersections = True
    """
    Here we define the endpoint of this tutorial, which is the CFFConverter node, as well as a few nodes which use
    the Nipype Merge utility. These are useful for passing lists of the files we want packaged in our CFF file.
    """

    CFFConverter = pe.Node(interface=cmtk.CFFConverter(), name="CFFConverter")

    giftiSurfaces = pe.Node(interface=util.Merge(8), name="GiftiSurfaces")
    giftiLabels = pe.Node(interface=util.Merge(2), name="GiftiLabels")
    niftiVolumes = pe.Node(interface=util.Merge(3), name="NiftiVolumes")
    fiberDataArrays = pe.Node(interface=util.Merge(4), name="FiberDataArrays")
    """
    Since we have now created all our nodes, we can define our workflow and start making connections.
    """

    mapping = pe.Workflow(name='mapping')
    """
    First, we connect the input node to the early conversion functions.
    FreeSurfer input nodes:
    """

    mapping.connect([(inputnode_within, FreeSurferSource, [("subjects_dir",
                                                            "subjects_dir")])])
    mapping.connect([(inputnode_within, FreeSurferSource, [("subject_id",
                                                            "subject_id")])])

    mapping.connect([(inputnode_within, FreeSurferSourceLH,
                      [("subjects_dir", "subjects_dir")])])
    mapping.connect([(inputnode_within, FreeSurferSourceLH, [("subject_id",
                                                              "subject_id")])])

    mapping.connect([(inputnode_within, FreeSurferSourceRH,
                      [("subjects_dir", "subjects_dir")])])
    mapping.connect([(inputnode_within, FreeSurferSourceRH, [("subject_id",
                                                              "subject_id")])])
    """
    Required conversions for processing in Camino:
    """

    mapping.connect([(inputnode_within, image2voxel,
                      [("dwi", "in_file")]), (inputnode_within, fsl2scheme,
                                              [("bvecs", "bvec_file"),
                                               ("bvals", "bval_file")]),
                     (image2voxel, dtifit, [['voxel_order', 'in_file']]),
                     (fsl2scheme, dtifit, [['scheme', 'scheme_file']])])
    """
    Nifti conversions for the subject's stripped brain image from Freesurfer:
    """

    mapping.connect([(FreeSurferSource, mri_convert_Brain, [('brain',
                                                             'in_file')])])
    """
    Surface conversions to GIFTI (pial, white, inflated, and sphere for both hemispheres)
    """

    mapping.connect([(FreeSurferSourceLH, mris_convertLH, [('pial',
                                                            'in_file')])])
    mapping.connect([(FreeSurferSourceRH, mris_convertRH, [('pial',
                                                            'in_file')])])
    mapping.connect([(FreeSurferSourceLH, mris_convertLHwhite, [('white',
                                                                 'in_file')])])
    mapping.connect([(FreeSurferSourceRH, mris_convertRHwhite, [('white',
                                                                 'in_file')])])
    mapping.connect([(FreeSurferSourceLH, mris_convertLHinflated,
                      [('inflated', 'in_file')])])
    mapping.connect([(FreeSurferSourceRH, mris_convertRHinflated,
                      [('inflated', 'in_file')])])
    mapping.connect([(FreeSurferSourceLH, mris_convertLHsphere,
                      [('sphere', 'in_file')])])
    mapping.connect([(FreeSurferSourceRH, mris_convertRHsphere,
                      [('sphere', 'in_file')])])
    """
    The annotation files are converted using the pial surface as a map via the MRIsConvert interface.
    One of the functions defined earlier is used to select the lh.aparc.annot and rh.aparc.annot files
    specifically (rather than i.e. rh.aparc.a2009s.annot) from the output list given by the FreeSurferSource.
    """

    mapping.connect([(FreeSurferSourceLH, mris_convertLHlabels,
                      [('pial', 'in_file')])])
    mapping.connect([(FreeSurferSourceRH, mris_convertRHlabels,
                      [('pial', 'in_file')])])
    mapping.connect([(FreeSurferSourceLH, mris_convertLHlabels,
                      [(('annot', select_aparc_annot), 'annot_file')])])
    mapping.connect([(FreeSurferSourceRH, mris_convertRHlabels,
                      [(('annot', select_aparc_annot), 'annot_file')])])
    """
    This section coregisters the diffusion-weighted and parcellated white-matter / whole brain images.
    At present the conmap node connection is left commented, as there have been recent changes in Camino
    code that have presented some users with errors.
    """

    mapping.connect([(inputnode_within, b0Strip, [('dwi', 'in_file')])])
    mapping.connect([(inputnode_within, b0Strip, [('dwi', 't2_guided')])
                     ])  # Added to improve damaged brain extraction
    mapping.connect([(b0Strip, coregister, [('out_file', 'in_file')])])
    mapping.connect([(mri_convert_Brain, coregister, [('out_file',
                                                       'reference')])])
    mapping.connect([(coregister, convertxfm, [('out_matrix_file',
                                                'in_file')])])
    mapping.connect([(b0Strip, inverse, [('out_file', 'reference')])])
    mapping.connect([(convertxfm, inverse, [('out_file', 'in_matrix_file')])])
    mapping.connect([(mri_convert_Brain, inverse, [('out_file', 'in_file')])])
    """
    The tractography pipeline consists of the following nodes. Further information about the tractography
    can be found in nipype/examples/dmri_camino_dti.py.
    """

    mapping.connect([(b0Strip, track, [("mask_file", "seed_file")])])
    mapping.connect([(fsl2scheme, dtlutgen, [("scheme", "scheme_file")])])
    mapping.connect([(dtlutgen, picopdfs, [("dtLUT", "luts")])])
    mapping.connect([(dtifit, picopdfs, [("tensor_fitted", "in_file")])])
    mapping.connect([(picopdfs, track, [("pdfs", "in_file")])])
    """
    Connecting the Fractional Anisotropy and Trace nodes is simple, as they obtain their input from the
    tensor fitting. This is also where our voxel- and data-grabbing functions come in. We pass these functions,
    along with the original DWI image from the input node, to the header-generating nodes. This ensures that the
    files will be correct and readable.
    """

    mapping.connect([(dtifit, fa, [("tensor_fitted", "in_file")])])
    mapping.connect([(fa, analyzeheader_fa, [("fa", "in_file")])])
    mapping.connect([(inputnode_within, analyzeheader_fa,
                      [(('dwi', get_vox_dims), 'voxel_dims'),
                       (('dwi', get_data_dims), 'data_dims')])])
    mapping.connect([(fa, fa2nii, [('fa', 'data_file')])])
    mapping.connect([(inputnode_within, fa2nii, [(('dwi', get_affine),
                                                  'affine')])])
    mapping.connect([(analyzeheader_fa, fa2nii, [('header', 'header_file')])])

    mapping.connect([(dtifit, trace, [("tensor_fitted", "in_file")])])
    mapping.connect([(trace, analyzeheader_trace, [("trace", "in_file")])])
    mapping.connect([(inputnode_within, analyzeheader_trace,
                      [(('dwi', get_vox_dims), 'voxel_dims'),
                       (('dwi', get_data_dims), 'data_dims')])])
    mapping.connect([(trace, trace2nii, [('trace', 'data_file')])])
    mapping.connect([(inputnode_within, trace2nii, [(('dwi', get_affine),
                                                     'affine')])])
    mapping.connect([(analyzeheader_trace, trace2nii, [('header',
                                                        'header_file')])])

    mapping.connect([(dtifit, dteig, [("tensor_fitted", "in_file")])])
    """
    The output tracts are converted to Trackvis format (and back). Here we also use the voxel- and data-grabbing
    functions defined at the beginning of the pipeline.
    """

    mapping.connect([(track, camino2trackvis, [('tracked', 'in_file')]),
                     (track, vtkstreamlines, [['tracked', 'in_file']]),
                     (camino2trackvis, trk2camino, [['trackvis', 'in_file']])])
    mapping.connect([(inputnode_within, camino2trackvis,
                      [(('dwi', get_vox_dims), 'voxel_dims'),
                       (('dwi', get_data_dims), 'data_dims')])])
    """
    Here the CMTK connectivity mapping nodes are connected.
    The original aparc+aseg image is converted to NIFTI, then registered to
    the diffusion image and delivered to the ROIGen node. The remapped parcellation,
    original tracts, and label file are then given to CreateMatrix.
    """

    mapping.connect(inputnode_within, 'resolution_network_file', createnodes,
                    'resolution_network_file')
    mapping.connect(createnodes, 'node_network', creatematrix,
                    'resolution_network_file')
    mapping.connect([(FreeSurferSource, mri_convert_AparcAseg,
                      [(('aparc_aseg', select_aparc), 'in_file')])])

    mapping.connect([(b0Strip, inverse_AparcAseg, [('out_file',
                                                    'reference')])])
    mapping.connect([(convertxfm, inverse_AparcAseg, [('out_file',
                                                       'in_matrix_file')])])
    mapping.connect([(mri_convert_AparcAseg, inverse_AparcAseg,
                      [('out_file', 'in_file')])])
    mapping.connect([(mri_convert_AparcAseg, roigen_structspace,
                      [('out_file', 'aparc_aseg_file')])])
    mapping.connect([(roigen_structspace, createnodes, [("roi_file",
                                                         "roi_file")])])

    mapping.connect([(inverse_AparcAseg, roigen, [("out_file",
                                                   "aparc_aseg_file")])])
    mapping.connect([(roigen, creatematrix, [("roi_file", "roi_file")])])
    mapping.connect([(camino2trackvis, creatematrix, [("trackvis",
                                                       "tract_file")])])
    mapping.connect([(inputnode_within, creatematrix, [("subject_id",
                                                        "out_matrix_file")])])
    mapping.connect([(inputnode_within, creatematrix,
                      [("subject_id", "out_matrix_mat_file")])])
    """
    The merge nodes defined earlier are used here to create lists of the files which are
    destined for the CFFConverter.
    """

    mapping.connect([(mris_convertLH, giftiSurfaces, [("converted", "in1")])])
    mapping.connect([(mris_convertRH, giftiSurfaces, [("converted", "in2")])])
    mapping.connect([(mris_convertLHwhite, giftiSurfaces, [("converted",
                                                            "in3")])])
    mapping.connect([(mris_convertRHwhite, giftiSurfaces, [("converted",
                                                            "in4")])])
    mapping.connect([(mris_convertLHinflated, giftiSurfaces, [("converted",
                                                               "in5")])])
    mapping.connect([(mris_convertRHinflated, giftiSurfaces, [("converted",
                                                               "in6")])])
    mapping.connect([(mris_convertLHsphere, giftiSurfaces, [("converted",
                                                             "in7")])])
    mapping.connect([(mris_convertRHsphere, giftiSurfaces, [("converted",
                                                             "in8")])])

    mapping.connect([(mris_convertLHlabels, giftiLabels, [("converted",
                                                           "in1")])])
    mapping.connect([(mris_convertRHlabels, giftiLabels, [("converted",
                                                           "in2")])])

    mapping.connect([(roigen, niftiVolumes, [("roi_file", "in1")])])
    mapping.connect([(inputnode_within, niftiVolumes, [("dwi", "in2")])])
    mapping.connect([(mri_convert_Brain, niftiVolumes, [("out_file", "in3")])])

    mapping.connect([(creatematrix, fiberDataArrays, [("endpoint_file",
                                                       "in1")])])
    mapping.connect([(creatematrix, fiberDataArrays, [("endpoint_file_mm",
                                                       "in2")])])
    mapping.connect([(creatematrix, fiberDataArrays, [("fiber_length_file",
                                                       "in3")])])
    mapping.connect([(creatematrix, fiberDataArrays, [("fiber_label_file",
                                                       "in4")])])
    """
    This block actually connects the merged lists to the CFF converter. We pass the surfaces
    and volumes that are to be included, as well as the tracts and the network itself. The currently
    running pipeline (dmri_connectivity.py) is also scraped and included in the CFF file. This
    makes it easy for the user to examine the entire processing pathway used to generate the end
    product.
    """

    CFFConverter.inputs.script_files = op.abspath(
        inspect.getfile(inspect.currentframe()))
    mapping.connect([(giftiSurfaces, CFFConverter, [("out",
                                                     "gifti_surfaces")])])
    mapping.connect([(giftiLabels, CFFConverter, [("out", "gifti_labels")])])
    mapping.connect([(creatematrix, CFFConverter, [("matrix_files",
                                                    "gpickled_networks")])])

    mapping.connect([(niftiVolumes, CFFConverter, [("out", "nifti_volumes")])])
    mapping.connect([(fiberDataArrays, CFFConverter, [("out", "data_files")])])
    mapping.connect([(camino2trackvis, CFFConverter, [("trackvis",
                                                       "tract_files")])])
    mapping.connect([(inputnode_within, CFFConverter, [("subject_id",
                                                        "title")])])
    """
    Finally, we create another higher-level workflow to connect our mapping workflow with the info and datagrabbing nodes
    declared at the beginning. Our tutorial can is now extensible to any arbitrary number of subjects by simply adding
    their names to the subject list and their data to the proper folders.
    """

    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            "subject_id", "dwi", "bvecs", "bvals", "subjects_dir",
            "resolution_network_file"
        ]),
        name="inputnode")

    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            "fa", "struct", "trace", "tracts", "connectome", "cmatrix",
            "networks", "rois", "mean_fiber_length", "fiber_length_std",
            "tensors"
        ]),
        name="outputnode")

    connectivity = pe.Workflow(name="connectivity")
    connectivity.base_output_dir = name

    connectivity.connect([
        (inputnode, mapping,
         [("dwi", "inputnode_within.dwi"), ("bvals", "inputnode_within.bvals"),
          ("bvecs", "inputnode_within.bvecs"), ("subject_id",
                                                "inputnode_within.subject_id"),
          ("subjects_dir", "inputnode_within.subjects_dir"),
          ("resolution_network_file",
           "inputnode_within.resolution_network_file")])
    ])

    connectivity.connect(
        [(mapping, outputnode,
          [("camino2trackvis.trackvis",
            "tracts"), ("CFFConverter.connectome_file", "connectome"),
           ("CreateMatrix.matrix_mat_file",
            "cmatrix"), ("CreateMatrix.mean_fiber_length_matrix_mat_file",
                         "mean_fiber_length"),
           ("CreateMatrix.fiber_length_std_matrix_mat_file",
            "fiber_length_std"), ("fa2nii.nifti_file",
                                  "fa"), ("CreateMatrix.matrix_files",
                                          "networks"), ("ROIGen.roi_file",
                                                        "rois"),
           ("mri_convert_Brain.out_file",
            "struct"), ("trace2nii.nifti_file",
                        "trace"), ("dtifit.tensor_fitted", "tensors")])])

    return connectivity



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function, unicode_literals
from future import standard_library
import os
import warnings

import pytest

from ....utils.filemanip import split_filename
from ... import base as nib
from ...base import traits, Undefined

standard_library.install_aliases()


@pytest.fixture(scope="module")
def setup_file(request, tmpdir_factory):
    tmp_dir = tmpdir_factory.mktemp('files')
    tmp_infile = tmp_dir.join('foo.txt')
    with tmp_infile.open('w') as fp:
        fp.writelines(['123456789'])

    tmp_dir.chdir()

    return tmp_infile.strpath


def test_TraitedSpec():
    assert nib.TraitedSpec().get_hashval()
    assert nib.TraitedSpec().__repr__() == '\n\n'

    class spec(nib.TraitedSpec):
        foo = nib.traits.Int
        goo = nib.traits.Float(usedefault=True)

    assert spec().foo == Undefined
    assert spec().goo == 0.0
    specfunc = lambda x: spec(hoo=x)
    with pytest.raises(nib.traits.TraitError):
        specfunc(1)
    infields = spec(foo=1)
    hashval = ([('foo', 1), ('goo', '0.0000000000')],
               'e89433b8c9141aa0fda2f8f4d662c047')
    assert infields.get_hashval() == hashval
    assert infields.__repr__() == '\nfoo = 1\ngoo = 0.0\n'


@pytest.mark.skip
def test_TraitedSpec_dynamic():
    from pickle import dumps, loads
    a = nib.BaseTraitedSpec()
    a.add_trait('foo', nib.traits.Int)
    a.foo = 1
    assign_a = lambda: setattr(a, 'foo', 'a')
    with pytest.raises(Exception):
        assign_a
    pkld_a = dumps(a)
    unpkld_a = loads(pkld_a)
    assign_a_again = lambda: setattr(unpkld_a, 'foo', 'a')
    with pytest.raises(Exception):
        assign_a_again


def test_TraitedSpec_logic():
    class spec3(nib.TraitedSpec):
        _xor_inputs = ('foo', 'bar')

        foo = nib.traits.Int(xor=_xor_inputs, desc='foo or bar, not both')
        bar = nib.traits.Int(xor=_xor_inputs, desc='bar or foo, not both')
        kung = nib.traits.Float(
            requires=('foo', ), position=0, desc='kung foo')

    class out3(nib.TraitedSpec):
        output = nib.traits.Int

    class MyInterface(nib.BaseInterface):
        input_spec = spec3
        output_spec = out3

    myif = MyInterface()
    # NOTE_dj, FAIL: I don't get a TypeError, only a UserWarning
    # with pytest.raises(TypeError):
    #     setattr(myif.inputs, 'kung', 10.0)
    myif.inputs.foo = 1
    assert myif.inputs.foo == 1
    set_bar = lambda: setattr(myif.inputs, 'bar', 1)
    with pytest.raises(IOError):
        set_bar()
    assert myif.inputs.foo == 1
    myif.inputs.kung = 2
    assert myif.inputs.kung == 2.0


def test_deprecation():
    with warnings.catch_warnings(record=True) as w:
        warnings.filterwarnings('always', '', UserWarning)

        class DeprecationSpec1(nib.TraitedSpec):
            foo = nib.traits.Int(deprecated='0.1')

        spec_instance = DeprecationSpec1()
        set_foo = lambda: setattr(spec_instance, 'foo', 1)
        with pytest.raises(nib.TraitError):
            set_foo()
        assert len(w) == 0, 'no warnings, just errors'

    with warnings.catch_warnings(record=True) as w:
        warnings.filterwarnings('always', '', UserWarning)

        class DeprecationSpec2(nib.TraitedSpec):
            foo = nib.traits.Int(deprecated='100', new_name='bar')

        spec_instance = DeprecationSpec2()
        set_foo = lambda: setattr(spec_instance, 'foo', 1)
        with pytest.raises(nib.TraitError):
            set_foo()
        assert len(w) == 0, 'no warnings, just errors'

    with warnings.catch_warnings(record=True) as w:
        warnings.filterwarnings('always', '', UserWarning)

        class DeprecationSpec3(nib.TraitedSpec):
            foo = nib.traits.Int(deprecated='1000', new_name='bar')
            bar = nib.traits.Int()

        spec_instance = DeprecationSpec3()
        not_raised = True
        try:
            spec_instance.foo = 1
        except nib.TraitError:
            not_raised = False
        assert not_raised
        assert len(
            w) == 1, 'deprecated warning 1 %s' % [w1.message for w1 in w]

    with warnings.catch_warnings(record=True) as w:
        warnings.filterwarnings('always', '', UserWarning)

        class DeprecationSpec3(nib.TraitedSpec):
            foo = nib.traits.Int(deprecated='1000', new_name='bar')
            bar = nib.traits.Int()

        spec_instance = DeprecationSpec3()
        not_raised = True
        try:
            spec_instance.foo = 1
        except nib.TraitError:
            not_raised = False
        assert not_raised
        assert spec_instance.foo == Undefined
        assert spec_instance.bar == 1
        assert len(
            w) == 1, 'deprecated warning 2 %s' % [w1.message for w1 in w]


def test_namesource(setup_file):
    tmp_infile = setup_file
    tmpd, nme, ext = split_filename(tmp_infile)

    class spec2(nib.CommandLineInputSpec):
        moo = nib.File(
            name_source=['doo'], hash_files=False, argstr="%s", position=2)
        doo = nib.File(exists=True, argstr="%s", position=1)
        goo = traits.Int(argstr="%d", position=4)
        poo = nib.File(
            name_source=['goo'], hash_files=False, argstr="%s", position=3)

    class TestName(nib.CommandLine):
        _cmd = "mycommand"
        input_spec = spec2

    testobj = TestName()
    testobj.inputs.doo = tmp_infile
    testobj.inputs.goo = 99
    assert '%s_generated' % nme in testobj.cmdline
    assert '%d_generated' % testobj.inputs.goo in testobj.cmdline
    testobj.inputs.moo = "my_%s_template"
    assert 'my_%s_template' % nme in testobj.cmdline


def test_chained_namesource(setup_file):
    tmp_infile = setup_file
    tmpd, nme, ext = split_filename(tmp_infile)

    class spec2(nib.CommandLineInputSpec):
        doo = nib.File(exists=True, argstr="%s", position=1)
        moo = nib.File(
            name_source=['doo'],
            hash_files=False,
            argstr="%s",
            position=2,
            name_template='%s_mootpl')
        poo = nib.File(
            name_source=['moo'], hash_files=False, argstr="%s", position=3)

    class TestName(nib.CommandLine):
        _cmd = "mycommand"
        input_spec = spec2

    testobj = TestName()
    testobj.inputs.doo = tmp_infile
    res = testobj.cmdline
    assert '%s' % tmp_infile in res
    assert '%s_mootpl ' % nme in res
    assert '%s_mootpl_generated' % nme in res


def test_cycle_namesource1(setup_file):
    tmp_infile = setup_file
    tmpd, nme, ext = split_filename(tmp_infile)

    class spec3(nib.CommandLineInputSpec):
        moo = nib.File(
            name_source=['doo'],
            hash_files=False,
            argstr="%s",
            position=1,
            name_template='%s_mootpl')
        poo = nib.File(
            name_source=['moo'], hash_files=False, argstr="%s", position=2)
        doo = nib.File(
            name_source=['poo'], hash_files=False, argstr="%s", position=3)

    class TestCycle(nib.CommandLine):
        _cmd = "mycommand"
        input_spec = spec3

    # Check that an exception is raised
    to0 = TestCycle()
    not_raised = True
    try:
        to0.cmdline
    except nib.NipypeInterfaceError:
        not_raised = False
    assert not not_raised


def test_cycle_namesource2(setup_file):
    tmp_infile = setup_file
    tmpd, nme, ext = split_filename(tmp_infile)

    class spec3(nib.CommandLineInputSpec):
        moo = nib.File(
            name_source=['doo'],
            hash_files=False,
            argstr="%s",
            position=1,
            name_template='%s_mootpl')
        poo = nib.File(
            name_source=['moo'], hash_files=False, argstr="%s", position=2)
        doo = nib.File(
            name_source=['poo'], hash_files=False, argstr="%s", position=3)

    class TestCycle(nib.CommandLine):
        _cmd = "mycommand"
        input_spec = spec3

    # Check that loop can be broken by setting one of the inputs
    to1 = TestCycle()
    to1.inputs.poo = tmp_infile

    not_raised = True
    try:
        res = to1.cmdline
    except nib.NipypeInterfaceError:
        not_raised = False
    print(res)

    assert not_raised
    assert '%s' % tmp_infile in res
    assert '%s_generated' % nme in res
    assert '%s_generated_mootpl' % nme in res


def test_namesource_constraints(setup_file):
    tmp_infile = setup_file
    tmpd, nme, ext = split_filename(tmp_infile)

    class constrained_spec(nib.CommandLineInputSpec):
        in_file = nib.File(argstr="%s", position=1)
        threshold = traits.Float(
            argstr="%g",
            xor=['mask_file'],
            position=2)
        mask_file = nib.File(
            argstr="%s",
            name_source=['in_file'],
            name_template='%s_mask',
            keep_extension=True,
            xor=['threshold'],
            position=2)
        out_file1 = nib.File(
            argstr="%s",
            name_source=['in_file'],
            name_template='%s_out1',
            keep_extension=True,
            position=3)
        out_file2 = nib.File(
            argstr="%s",
            name_source=['in_file'],
            name_template='%s_out2',
            keep_extension=True,
            requires=['threshold'],
            position=4)

    class TestConstrained(nib.CommandLine):
        _cmd = "mycommand"
        input_spec = constrained_spec

    tc = TestConstrained()

    # name_source undefined, so template traits remain undefined
    assert tc.cmdline == 'mycommand'

    # mask_file and out_file1 enabled by name_source definition
    tc.inputs.in_file = os.path.basename(tmp_infile)
    assert tc.cmdline == 'mycommand foo.txt foo_mask.txt foo_out1.txt'

    # mask_file disabled by threshold, out_file2 enabled by threshold
    tc.inputs.threshold = 10.
    assert tc.cmdline == 'mycommand foo.txt 10 foo_out1.txt foo_out2.txt'


def test_TraitedSpec_withFile(setup_file):
    tmp_infile = setup_file
    tmpd, nme = os.path.split(tmp_infile)
    assert os.path.exists(tmp_infile)

    class spec2(nib.TraitedSpec):
        moo = nib.File(exists=True)
        doo = nib.traits.List(nib.File(exists=True))

    infields = spec2(moo=tmp_infile, doo=[tmp_infile])
    hashval = infields.get_hashval(hash_method='content')
    assert hashval[1] == 'a00e9ee24f5bfa9545a515b7a759886b'


def test_TraitedSpec_withNoFileHashing(setup_file):
    tmp_infile = setup_file
    tmpd, nme = os.path.split(tmp_infile)
    assert os.path.exists(tmp_infile)

    class spec2(nib.TraitedSpec):
        moo = nib.File(exists=True, hash_files=False)
        doo = nib.traits.List(nib.File(exists=True))

    infields = spec2(moo=nme, doo=[tmp_infile])
    hashval = infields.get_hashval(hash_method='content')
    assert hashval[1] == '8da4669ff5d72f670a46ea3e7a203215'

    class spec3(nib.TraitedSpec):
        moo = nib.File(exists=True, name_source="doo")
        doo = nib.traits.List(nib.File(exists=True))

    infields = spec3(moo=nme, doo=[tmp_infile])
    hashval1 = infields.get_hashval(hash_method='content')

    class spec4(nib.TraitedSpec):
        moo = nib.File(exists=True)
        doo = nib.traits.List(nib.File(exists=True))

    infields = spec4(moo=nme, doo=[tmp_infile])
    hashval2 = infields.get_hashval(hash_method='content')
    assert hashval1[1] != hashval2[1]


def test_ImageFile():
    x = nib.BaseInterface().inputs

    # setup traits
    x.add_trait('nifti', nib.ImageFile(types=['nifti1', 'dicom']))
    x.add_trait('anytype', nib.ImageFile())
    x.add_trait('newtype', nib.ImageFile(types=['nifti10']))
    x.add_trait('nocompress',
                nib.ImageFile(types=['mgh'], allow_compressed=False))

    with pytest.raises(nib.TraitError):
        x.nifti = 'test.mgz'
    x.nifti = 'test.nii'
    x.anytype = 'test.xml'
    with pytest.raises(AttributeError):
        x.newtype = 'test.nii'
    with pytest.raises(nib.TraitError):
        x.nocompress = 'test.nii.gz'
    x.nocompress = 'test.mgh'



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..gtract import gtractResampleDWIInPlace


def test_gtractResampleDWIInPlace_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        debugLevel=dict(argstr='--debugLevel %d', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        imageOutputSize=dict(
            argstr='--imageOutputSize %s',
            sep=',',
        ),
        inputTransform=dict(argstr='--inputTransform %s', ),
        inputVolume=dict(argstr='--inputVolume %s', ),
        numberOfThreads=dict(argstr='--numberOfThreads %d', ),
        outputResampledB0=dict(
            argstr='--outputResampledB0 %s',
            hash_files=False,
        ),
        outputVolume=dict(
            argstr='--outputVolume %s',
            hash_files=False,
        ),
        referenceVolume=dict(argstr='--referenceVolume %s', ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
        warpDWITransform=dict(argstr='--warpDWITransform %s', ),
    )
    inputs = gtractResampleDWIInPlace.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_gtractResampleDWIInPlace_outputs():
    output_map = dict(
        outputResampledB0=dict(),
        outputVolume=dict(),
    )
    outputs = gtractResampleDWIInPlace.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



""" This file contains defines parameters for nipy that we use to fill
settings in setup.py, the nipy top-level docstring, and for building the
docs.  In setup.py in particular, we exec this file, so it cannot import nipy
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import sys

# nipype version information.  An empty version_extra corresponds to a
# full release.  '.dev' as a version_extra string means this is a development
# version
# Remove -dev for release
__version__ = '1.0.3'


def get_nipype_gitversion():
    """Nipype version as reported by the last commit in git

    Returns
    -------
    None or str
      Version of Nipype according to git.
    """
    import os
    import subprocess
    try:
        import nipype
        gitpath = os.path.realpath(
            os.path.join(os.path.dirname(nipype.__file__), os.path.pardir))
    except:
        gitpath = os.getcwd()
    gitpathgit = os.path.join(gitpath, '.git')
    if not os.path.exists(gitpathgit):
        return None
    ver = None
    try:
        o, _ = subprocess.Popen(
            'git describe', shell=True, cwd=gitpath,
            stdout=subprocess.PIPE).communicate()
    except Exception:
        pass
    else:
        ver = o.decode().strip().split('-')[-1]
    return ver


if __version__.endswith('-dev'):
    gitversion = get_nipype_gitversion()
    if gitversion:
        __version__ = '{}+{}'.format(__version__, gitversion)

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable', 'Environment :: Console',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6', 'Topic :: Scientific/Engineering'
]

description = 'Neuroimaging in Python: Pipelines and Interfaces'

# Note: this long_description is actually a copy/paste from the top-level
# README.txt, so that it shows up nicely on PyPI.  So please remember to edit
# it only in one place and sync it correctly.
long_description = """========================================================
NIPYPE: Neuroimaging in Python: Pipelines and Interfaces
========================================================

Current neuroimaging software offer users an incredible opportunity to \
analyze data using a variety of different algorithms. However, this has \
resulted in a heterogeneous collection of specialized applications \
without transparent interoperability or a uniform operating interface.

*Nipype*, an open-source, community-developed initiative under the \
umbrella of NiPy_, is a Python project that provides a uniform interface \
to existing neuroimaging software and facilitates interaction between \
these packages within a single workflow. Nipype provides an environment \
that encourages interactive exploration of algorithms from different \
packages (e.g., AFNI, ANTS, BRAINS, BrainSuite, Camino, FreeSurfer, FSL, MNE, \
MRtrix, MNE, Nipy, Slicer, SPM), eases the design of workflows within and \
between packages, and reduces the learning curve necessary to use different \
packages. Nipype is creating a collaborative platform for neuroimaging \
software development in a high-level language and addressing limitations of \
existing pipeline systems.

*Nipype* allows you to:

* easily interact with tools from different software packages
* combine processing steps from different software packages
* develop new workflows faster by reusing common steps from old ones
* process data faster by running it in parallel on many cores/machines
* make your research easily reproducible
* share your processing workflows with the community
"""

# versions
NIBABEL_MIN_VERSION = '2.1.0'
NETWORKX_MIN_VERSION = '1.9'
NUMPY_MIN_VERSION = '1.9.0'
SCIPY_MIN_VERSION = '0.14'
TRAITS_MIN_VERSION = '4.6'
DATEUTIL_MIN_VERSION = '2.2'
PYTEST_MIN_VERSION = '3.0'
FUTURE_MIN_VERSION = '0.16.0'
SIMPLEJSON_MIN_VERSION = '3.8.0'
PROV_VERSION = '1.5.0'
CLICK_MIN_VERSION = '6.6.0'
PYDOT_MIN_VERSION = '1.2.3'

NAME = 'nipype'
MAINTAINER = 'nipype developers'
MAINTAINER_EMAIL = 'neuroimaging@python.org'
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = 'http://nipy.org/nipype'
DOWNLOAD_URL = 'http://github.com/nipy/nipype/archives/master'
LICENSE = 'Apache License, 2.0'
AUTHOR = 'nipype developers'
AUTHOR_EMAIL = 'neuroimaging@python.org'
PLATFORMS = 'OS Independent'
MAJOR = __version__.split('.')[0]
MINOR = __version__.split('.')[1]
MICRO = __version__.replace('-', '.').split('.')[2]
ISRELEASE = (len(__version__.replace('-', '.').split('.')) == 3
             or 'post' in __version__.replace('-', '.').split('.')[-1])
VERSION = __version__
PROVIDES = ['nipype']
REQUIRES = [
    'nibabel>=%s' % NIBABEL_MIN_VERSION,
    'networkx>=%s' % NETWORKX_MIN_VERSION,
    'numpy>=%s' % NUMPY_MIN_VERSION,
    'python-dateutil>=%s' % DATEUTIL_MIN_VERSION,
    'scipy>=%s' % SCIPY_MIN_VERSION,
    'traits>=%s' % TRAITS_MIN_VERSION,
    'future>=%s' % FUTURE_MIN_VERSION,
    'simplejson>=%s' % SIMPLEJSON_MIN_VERSION,
    'prov==%s' % PROV_VERSION,
    'click>=%s' % CLICK_MIN_VERSION,
    'funcsigs',
    'pytest>=%s' % PYTEST_MIN_VERSION,
    'mock',
    'pydotplus',
    'pydot>=%s' % PYDOT_MIN_VERSION,
    'packaging',
]

if sys.version_info <= (3, 4):
    REQUIRES.append('configparser')

TESTS_REQUIRES = ['pytest-cov', 'codecov', 'pytest-xdist', 'pytest-env']

EXTRA_REQUIRES = {
    'doc': ['Sphinx>=1.4', 'numpydoc', 'matplotlib', 'pydotplus', 'pydot>=1.2.3'],
    'tests': TESTS_REQUIRES,
    'specs': ['yapf'],
    'nipy': ['nitime', 'nilearn', 'dipy', 'nipy', 'matplotlib'],
    'profiler': ['psutil>=5.0'],
    'duecredit': ['duecredit'],
    'xvfbwrapper': ['xvfbwrapper'],
    'pybids': ['pybids'],
    'ssh': ['paramiko'],
    # 'mesh': ['mayavi']  # Enable when it works
}


def _list_union(iterable):
    return list(set(sum(iterable, [])))


# Enable a handle to install all extra dependencies at once
EXTRA_REQUIRES['all'] = _list_union(EXTRA_REQUIRES.values())
# dev = doc + tests + specs
EXTRA_REQUIRES['dev'] = _list_union(val for key, val in EXTRA_REQUIRES.items()
                                    if key in ('doc', 'tests', 'specs'))

STATUS = 'stable'



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from .nx import (create_networkx_pipeline, create_cmats_to_csv_pipeline)
from .group_connectivity import (
    create_merge_networks_by_group_workflow,
    create_merge_network_results_by_group_workflow,
    create_merge_group_networks_workflow,
    create_merge_group_network_results_workflow,
    create_average_networks_by_group_workflow)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import os

from ....algorithms import rapidart as ra
from ....interfaces import spm as spm
from ....interfaces import utility as niu
from ....pipeline import engine as pe
from ...smri.freesurfer.utils import create_getmask_flow

from .... import logging
logger = logging.getLogger('workflow')


def create_spm_preproc(name='preproc'):
    """Create an spm preprocessing workflow with freesurfer registration and
    artifact detection.

    The workflow realigns and smooths and registers the functional images with
    the subject's freesurfer space.

    Example
    -------

    >>> preproc = create_spm_preproc()
    >>> preproc.base_dir = '.'
    >>> preproc.inputs.inputspec.fwhm = 6
    >>> preproc.inputs.inputspec.subject_id = 's1'
    >>> preproc.inputs.inputspec.subjects_dir = '.'
    >>> preproc.inputs.inputspec.functionals = ['f3.nii', 'f5.nii']
    >>> preproc.inputs.inputspec.norm_threshold = 1
    >>> preproc.inputs.inputspec.zintensity_threshold = 3

    Inputs::

         inputspec.functionals : functional runs use 4d nifti
         inputspec.subject_id : freesurfer subject id
         inputspec.subjects_dir : freesurfer subjects dir
         inputspec.fwhm : smoothing fwhm
         inputspec.norm_threshold : norm threshold for outliers
         inputspec.zintensity_threshold : intensity threshold in z-score

    Outputs::

         outputspec.realignment_parameters : realignment parameter files
         outputspec.smoothed_files : smoothed functional files
         outputspec.outlier_files : list of outliers
         outputspec.outlier_stats : statistics of outliers
         outputspec.outlier_plots : images of outliers
         outputspec.mask_file : binary mask file in reference image space
         outputspec.reg_file : registration file that maps reference image to
                                 freesurfer space
         outputspec.reg_cost : cost of registration (useful for detecting misalignment)
    """
    """
    Initialize the workflow
    """

    workflow = pe.Workflow(name=name)
    """
    Define the inputs to this workflow
    """

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'functionals', 'subject_id', 'subjects_dir', 'fwhm',
            'norm_threshold', 'zintensity_threshold'
        ]),
        name='inputspec')
    """
    Setup the processing nodes and create the mask generation and coregistration
    workflow
    """

    poplist = lambda x: x.pop()
    realign = pe.Node(spm.Realign(), name='realign')
    workflow.connect(inputnode, 'functionals', realign, 'in_files')
    maskflow = create_getmask_flow()
    workflow.connect([(inputnode, maskflow,
                       [('subject_id', 'inputspec.subject_id'),
                        ('subjects_dir', 'inputspec.subjects_dir')])])
    maskflow.inputs.inputspec.contrast_type = 't2'
    workflow.connect(realign, 'mean_image', maskflow, 'inputspec.source_file')
    smooth = pe.Node(spm.Smooth(), name='smooth')
    workflow.connect(inputnode, 'fwhm', smooth, 'fwhm')
    workflow.connect(realign, 'realigned_files', smooth, 'in_files')
    artdetect = pe.Node(
        ra.ArtifactDetect(
            mask_type='file',
            parameter_source='SPM',
            use_differences=[True, False],
            use_norm=True,
            save_plot=True),
        name='artdetect')
    workflow.connect([(inputnode, artdetect,
                       [('norm_threshold', 'norm_threshold'),
                        ('zintensity_threshold', 'zintensity_threshold')])])
    workflow.connect([(realign, artdetect, [('realigned_files',
                                             'realigned_files'),
                                            ('realignment_parameters',
                                             'realignment_parameters')])])
    workflow.connect(maskflow, ('outputspec.mask_file', poplist), artdetect,
                     'mask_file')
    """
    Define the outputs of the workflow and connect the nodes to the outputnode
    """

    outputnode = pe.Node(
        niu.IdentityInterface(fields=[
            "realignment_parameters", "smoothed_files", "mask_file",
            "reg_file", "reg_cost", 'outlier_files', 'outlier_stats',
            'outlier_plots'
        ]),
        name="outputspec")
    workflow.connect(
        [(maskflow, outputnode, [("outputspec.reg_file", "reg_file")]),
         (maskflow, outputnode,
          [("outputspec.reg_cost", "reg_cost")]), (maskflow, outputnode, [
              (("outputspec.mask_file", poplist), "mask_file")
          ]), (realign, outputnode, [('realignment_parameters',
                                      'realignment_parameters')]),
         (smooth, outputnode, [('smoothed_files', 'smoothed_files')]),
         (artdetect, outputnode, [('outlier_files', 'outlier_files'),
                                  ('statistic_files', 'outlier_stats'),
                                  ('plot_files', 'outlier_plots')])])
    return workflow


def create_vbm_preproc(name='vbmpreproc'):
    """Create a vbm workflow that generates DARTEL-based warps to MNI space

    Based on: http://www.fil.ion.ucl.ac.uk/~john/misc/VBMclass10.pdf

    Example
    -------

    >>> preproc = create_vbm_preproc()
    >>> preproc.inputs.inputspec.fwhm = 8
    >>> preproc.inputs.inputspec.structural_files = [
    ...     os.path.abspath('s1.nii'), os.path.abspath('s3.nii')]
    >>> preproc.inputs.inputspec.template_prefix = 'Template'
    >>> preproc.run() # doctest: +SKIP

    Inputs::

         inputspec.structural_files : structural data to be used to create templates
         inputspec.fwhm: single of triplet for smoothing when normalizing to MNI space
         inputspec.template_prefix : prefix for dartel template

    Outputs::

         outputspec.normalized_files : normalized gray matter files
         outputspec.template_file : DARTEL template
         outputspec.icv : intracranial volume (cc - assuming dimensions in mm)

    """

    workflow = pe.Workflow(name=name)
    """
    Define the inputs to this workflow
    """

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['structural_files', 'fwhm', 'template_prefix']),
        name='inputspec')

    dartel_template = create_DARTEL_template()

    workflow.connect(inputnode, 'template_prefix', dartel_template,
                     'inputspec.template_prefix')
    workflow.connect(inputnode, 'structural_files', dartel_template,
                     'inputspec.structural_files')

    norm2mni = pe.Node(spm.DARTELNorm2MNI(modulate=True), name='norm2mni')
    workflow.connect(dartel_template, 'outputspec.template_file', norm2mni,
                     'template_file')
    workflow.connect(dartel_template, 'outputspec.flow_fields', norm2mni,
                     'flowfield_files')

    def getclass1images(class_images):
        class1images = []
        for session in class_images:
            class1images.extend(session[0])
        return class1images

    workflow.connect(dartel_template,
                     ('segment.native_class_images', getclass1images),
                     norm2mni, 'apply_to_files')
    workflow.connect(inputnode, 'fwhm', norm2mni, 'fwhm')

    def compute_icv(class_images):
        from nibabel import load
        from numpy import prod
        icv = []
        for session in class_images:
            voxel_volume = prod(load(session[0][0]).header.get_zooms())
            img = load(session[0][0]).get_data() + \
                load(session[1][0]).get_data() + \
                load(session[2][0]).get_data()
            img_icv = (img > 0.5).astype(int).sum() * voxel_volume * 1e-3
            icv.append(img_icv)
        return icv

    calc_icv = pe.Node(
        niu.Function(
            function=compute_icv,
            input_names=['class_images'],
            output_names=['icv']),
        name='calc_icv')

    workflow.connect(dartel_template, 'segment.native_class_images', calc_icv,
                     'class_images')
    """
    Define the outputs of the workflow and connect the nodes to the outputnode
    """

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=["normalized_files", "template_file", "icv"]),
        name="outputspec")
    workflow.connect([
        (dartel_template, outputnode, [('outputspec.template_file',
                                        'template_file')]),
        (norm2mni, outputnode, [("normalized_files", "normalized_files")]),
        (calc_icv, outputnode, [("icv", "icv")]),
    ])

    return workflow


def create_DARTEL_template(name='dartel_template'):
    """Create a vbm workflow that generates DARTEL-based template


    Example
    -------

    >>> preproc = create_DARTEL_template()
    >>> preproc.inputs.inputspec.structural_files = [
    ...     os.path.abspath('s1.nii'), os.path.abspath('s3.nii')]
    >>> preproc.inputs.inputspec.template_prefix = 'Template'
    >>> preproc.run() # doctest: +SKIP

    Inputs::

         inputspec.structural_files : structural data to be used to create templates
         inputspec.template_prefix : prefix for dartel template

    Outputs::

         outputspec.template_file : DARTEL template
         outputspec.flow_fields : warps from input struct files to the template

    """

    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['structural_files', 'template_prefix']),
        name='inputspec')

    segment = pe.MapNode(
        spm.NewSegment(), iterfield=['channel_files'], name='segment')
    workflow.connect(inputnode, 'structural_files', segment, 'channel_files')

    spm_info = spm.Info.getinfo()
    if spm_info:
        spm_path = spm_info['path']
        if spm_info['name'] == 'SPM8':
            tissue1 = ((os.path.join(spm_path, 'toolbox/Seg/TPM.nii'), 1), 2,
                       (True, True), (False, False))
            tissue2 = ((os.path.join(spm_path, 'toolbox/Seg/TPM.nii'), 2), 2,
                       (True, True), (False, False))
            tissue3 = ((os.path.join(spm_path, 'toolbox/Seg/TPM.nii'), 3), 2,
                       (True, False), (False, False))
            tissue4 = ((os.path.join(spm_path, 'toolbox/Seg/TPM.nii'), 4), 3,
                       (False, False), (False, False))
            tissue5 = ((os.path.join(spm_path, 'toolbox/Seg/TPM.nii'), 5), 4,
                       (False, False), (False, False))
            tissue6 = ((os.path.join(spm_path, 'toolbox/Seg/TPM.nii'), 6), 2,
                       (False, False), (False, False))
        elif spm_info['name'] == 'SPM12':
            spm_path = spm_info['path']
            tissue1 = ((os.path.join(spm_path, 'tpm/TPM.nii'), 1), 1,
                       (True, True), (False, False))
            tissue2 = ((os.path.join(spm_path, 'tpm/TPM.nii'), 2), 1,
                       (True, True), (False, False))
            tissue3 = ((os.path.join(spm_path, 'tpm/TPM.nii'), 3), 2,
                       (True, False), (False, False))
            tissue4 = ((os.path.join(spm_path, 'tpm/TPM.nii'), 4), 3,
                       (False, False), (False, False))
            tissue5 = ((os.path.join(spm_path, 'tpm/TPM.nii'), 5), 4,
                       (False, False), (False, False))
            tissue6 = ((os.path.join(spm_path, 'tpm/TPM.nii'), 6), 2,
                       (False, False), (False, False))
        else:
            logger.critical('Unsupported version of SPM')

        segment.inputs.tissues = [
            tissue1, tissue2, tissue3, tissue4, tissue5, tissue6
        ]
    else:
        logger.critical('SPM not found')

    dartel = pe.Node(spm.DARTEL(), name='dartel')
    """Get the gray and white segmentation classes generated by NewSegment
    """

    def get2classes(dartel_files):
        class1images = []
        class2images = []
        for session in dartel_files:
            class1images.extend(session[0])
            class2images.extend(session[1])
        return [class1images, class2images]

    workflow.connect(segment, ('dartel_input_images', get2classes), dartel,
                     'image_files')
    workflow.connect(inputnode, 'template_prefix', dartel, 'template_prefix')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=["template_file", "flow_fields"]),
        name="outputspec")
    workflow.connect([
        (dartel, outputnode, [('final_template_file', 'template_file'),
                              ('dartel_flow_fields', 'flow_fields')]),
    ])

    return workflow



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..nx import AverageNetworks


def test_AverageNetworks_inputs():
    input_map = dict(
        group_id=dict(usedefault=True, ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_files=dict(mandatory=True, ),
        out_gexf_groupavg=dict(),
        out_gpickled_groupavg=dict(),
        resolution_network_file=dict(),
    )
    inputs = AverageNetworks.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_AverageNetworks_outputs():
    output_map = dict(
        gexf_groupavg=dict(),
        gpickled_groupavg=dict(),
        matlab_groupavgs=dict(),
    )
    outputs = AverageNetworks.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..minc import Math


def test_Math_inputs():
    input_map = dict(
        abs=dict(argstr='-abs', ),
        args=dict(argstr='%s', ),
        calc_add=dict(argstr='-add', ),
        calc_and=dict(argstr='-and', ),
        calc_div=dict(argstr='-div', ),
        calc_mul=dict(argstr='-mult', ),
        calc_not=dict(argstr='-not', ),
        calc_or=dict(argstr='-or', ),
        calc_sub=dict(argstr='-sub', ),
        check_dimensions=dict(
            argstr='-check_dimensions',
            xor=('check_dimensions', 'no_check_dimensions'),
        ),
        clamp=dict(argstr='-clamp -const2 %s %s', ),
        clobber=dict(
            argstr='-clobber',
            usedefault=True,
        ),
        copy_header=dict(
            argstr='-copy_header',
            xor=('copy_header', 'no_copy_header'),
        ),
        count_valid=dict(argstr='-count_valid', ),
        dimension=dict(argstr='-dimension %s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        exp=dict(argstr='-exp -const2 %s %s', ),
        filelist=dict(
            argstr='-filelist %s',
            mandatory=True,
            xor=('input_files', 'filelist'),
        ),
        format_byte=dict(
            argstr='-byte',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_double=dict(
            argstr='-double',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_filetype=dict(
            argstr='-filetype',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_float=dict(
            argstr='-float',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_int=dict(
            argstr='-int',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_long=dict(
            argstr='-long',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_short=dict(
            argstr='-short',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_signed=dict(
            argstr='-signed',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        format_unsigned=dict(
            argstr='-unsigned',
            xor=('format_filetype', 'format_byte', 'format_short',
                 'format_int', 'format_long', 'format_float', 'format_double',
                 'format_signed', 'format_unsigned'),
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        ignore_nan=dict(argstr='-ignore_nan', ),
        input_files=dict(
            argstr='%s',
            mandatory=True,
            position=-2,
            sep=' ',
            xor=('input_files', 'filelist'),
        ),
        invert=dict(argstr='-invert -const %s', ),
        isnan=dict(argstr='-isnan', ),
        log=dict(argstr='-log -const2 %s %s', ),
        max_buffer_size_in_kb=dict(
            argstr='-max_buffer_size_in_kb %d',
            usedefault=True,
        ),
        maximum=dict(argstr='-maximum', ),
        minimum=dict(argstr='-minimum', ),
        nisnan=dict(argstr='-nisnan', ),
        no_check_dimensions=dict(
            argstr='-nocheck_dimensions',
            xor=('check_dimensions', 'no_check_dimensions'),
        ),
        no_copy_header=dict(
            argstr='-nocopy_header',
            xor=('copy_header', 'no_copy_header'),
        ),
        nsegment=dict(argstr='-nsegment -const2 %s %s', ),
        output_file=dict(
            argstr='%s',
            genfile=True,
            hash_files=False,
            name_source=['input_files'],
            name_template='%s_mincmath.mnc',
            position=-1,
        ),
        output_illegal=dict(
            argstr='-illegal_value',
            xor=('output_nan', 'output_zero', 'output_illegal_value'),
        ),
        output_nan=dict(
            argstr='-nan',
            xor=('output_nan', 'output_zero', 'output_illegal_value'),
        ),
        output_zero=dict(
            argstr='-zero',
            xor=('output_nan', 'output_zero', 'output_illegal_value'),
        ),
        percentdiff=dict(argstr='-percentdiff', ),
        propagate_nan=dict(argstr='-propagate_nan', ),
        scale=dict(argstr='-scale -const2 %s %s', ),
        segment=dict(argstr='-segment -const2 %s %s', ),
        sqrt=dict(argstr='-sqrt', ),
        square=dict(argstr='-square', ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
        test_eq=dict(argstr='-eq', ),
        test_ge=dict(argstr='-ge', ),
        test_gt=dict(argstr='-gt', ),
        test_le=dict(argstr='-le', ),
        test_lt=dict(argstr='-lt', ),
        test_ne=dict(argstr='-ne', ),
        two=dict(argstr='-2', ),
        voxel_range=dict(argstr='-range %d %d', ),
    )
    inputs = Math.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Math_outputs():
    output_map = dict(output_file=dict(), )
    outputs = Math.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""The dcm2nii module provides basic functions for dicom conversion
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import str, open
import os
import re
from copy import deepcopy

from ..utils.filemanip import split_filename
from .base import (CommandLine, CommandLineInputSpec, InputMultiPath, traits,
                   TraitedSpec, OutputMultiPath, isdefined, File, Directory,
                   PackageInfo)


class Info(PackageInfo):
    """Handle dcm2niix version information"""

    version_cmd = 'dcm2niix'

    @staticmethod
    def parse_version(raw_info):
        m = re.search(r'version (\S+)', raw_info)
        return m.groups()[0] if m else None


class Dcm2niiInputSpec(CommandLineInputSpec):
    source_names = InputMultiPath(
        File(exists=True),
        argstr="%s",
        position=-1,
        copyfile=False,
        mandatory=True,
        xor=['source_dir'])
    source_dir = Directory(
        exists=True,
        argstr="%s",
        position=-1,
        mandatory=True,
        xor=['source_names'])
    anonymize = traits.Bool(
        True,
        argstr='-a',
        usedefault=True,
        desc="Remove identifying information")
    config_file = File(
        exists=True,
        argstr="-b %s",
        genfile=True,
        desc="Load settings from specified inifile")
    collapse_folders = traits.Bool(
        True, argstr='-c', usedefault=True, desc="Collapse input folders")
    date_in_filename = traits.Bool(
        True, argstr='-d', usedefault=True, desc="Date in filename")
    events_in_filename = traits.Bool(
        True,
        argstr='-e',
        usedefault=True,
        desc="Events (series/acq) in filename")
    source_in_filename = traits.Bool(
        False, argstr='-f', usedefault=True, desc="Source filename")
    gzip_output = traits.Bool(
        False, argstr='-g', usedefault=True, desc="Gzip output (.gz)")
    id_in_filename = traits.Bool(
        False, argstr='-i', usedefault=True, desc="ID  in filename")
    nii_output = traits.Bool(
        True,
        argstr='-n',
        usedefault=True,
        desc="Save as .nii - if no, create .hdr/.img pair")
    output_dir = Directory(
        exists=True,
        argstr='-o %s',
        genfile=True,
        desc="Output dir - if unspecified, source directory is used")
    protocol_in_filename = traits.Bool(
        True, argstr='-p', usedefault=True, desc="Protocol in filename")
    reorient = traits.Bool(
        argstr='-r', desc="Reorient image to nearest orthogonal")
    spm_analyze = traits.Bool(
        argstr='-s', xor=['nii_output'], desc="SPM2/Analyze not SPM5/NIfTI")
    convert_all_pars = traits.Bool(
        True,
        argstr='-v',
        usedefault=True,
        desc="Convert every image in directory")
    reorient_and_crop = traits.Bool(
        False,
        argstr='-x',
        usedefault=True,
        desc="Reorient and crop 3D images")


class Dcm2niiOutputSpec(TraitedSpec):
    converted_files = OutputMultiPath(File(exists=True))
    reoriented_files = OutputMultiPath(File(exists=True))
    reoriented_and_cropped_files = OutputMultiPath(File(exists=True))
    bvecs = OutputMultiPath(File(exists=True))
    bvals = OutputMultiPath(File(exists=True))


class Dcm2nii(CommandLine):
    """Uses MRIcron's dcm2nii to convert dicom files

    Examples
    ========

    .. testsetup::

    >>> tmp = getfixture('tmpdir')
    >>> old = tmp.chdir() # changing to a temporary directory

    .. doctest::

    >>> from nipype.interfaces.dcm2nii import Dcm2nii
    >>> converter = Dcm2nii()
    >>> converter.inputs.source_names = [os.path.join(datadir, 'functional_1.dcm'), os.path.join(datadir, 'functional_2.dcm')]
    >>> converter.inputs.gzip_output = True
    >>> converter.inputs.output_dir = '.'
    >>> converter.cmdline #doctest: +ELLIPSIS
    'dcm2nii -a y -c y -b config.ini -v y -d y -e y -g y -i n -n y -o . -p y -x n -f n ...functional_1.dcm'

    .. testsetup::

    >>> os.chdir(old.strpath)

    """

    input_spec = Dcm2niiInputSpec
    output_spec = Dcm2niiOutputSpec
    _cmd = 'dcm2nii'

    def _format_arg(self, opt, spec, val):
        if opt in [
                'anonymize', 'collapse_folders', 'date_in_filename',
                'events_in_filename', 'source_in_filename', 'gzip_output',
                'id_in_filename', 'nii_output', 'protocol_in_filename',
                'reorient', 'spm_analyze', 'convert_all_pars',
                'reorient_and_crop'
        ]:
            spec = deepcopy(spec)
            if val:
                spec.argstr += ' y'
            else:
                spec.argstr += ' n'
                val = True
        if opt == 'source_names':
            return spec.argstr % val[0]
        return super(Dcm2nii, self)._format_arg(opt, spec, val)

    def _run_interface(self, runtime):
        self._config_created = False
        new_runtime = super(Dcm2nii, self)._run_interface(runtime)
        (self.output_files, self.reoriented_files,
         self.reoriented_and_cropped_files, self.bvecs,
         self.bvals) = self._parse_stdout(new_runtime.stdout)
        if self._config_created:
            os.remove('config.ini')
        return new_runtime

    def _parse_stdout(self, stdout):
        files = []
        reoriented_files = []
        reoriented_and_cropped_files = []
        bvecs = []
        bvals = []
        skip = False
        last_added_file = None
        for line in stdout.split("\n"):
            if not skip:
                out_file = None
                if line.startswith("Saving "):
                    out_file = line[len("Saving "):]
                elif line.startswith("GZip..."):
                    # for gzipped output files are not absolute
                    fname = line[len("GZip..."):]
                    if len(files) and os.path.basename(
                            files[-1]) == fname[:-3]:
                        # we are seeing a previously reported conversion
                        # as being saved in gzipped form -- remove the
                        # obsolete, uncompressed file
                        files.pop()
                    if isdefined(self.inputs.output_dir):
                        output_dir = self.inputs.output_dir
                    else:
                        output_dir = self._gen_filename('output_dir')
                    out_file = os.path.abspath(os.path.join(output_dir, fname))
                elif line.startswith("Number of diffusion directions "):
                    if last_added_file:
                        base, filename, ext = split_filename(last_added_file)
                        bvecs.append(os.path.join(base, filename + ".bvec"))
                        bvals.append(os.path.join(base, filename + ".bval"))
                elif line.startswith("Removed DWI from DTI scan"):
                    # such line can only follow the 'diffusion' case handled
                    # just above
                    for l in (bvecs, bvals):
                        l[-1] = os.path.join(
                            os.path.dirname(l[-1]),
                            'x%s' % (os.path.basename(l[-1]), ))
                elif re.search('.*->(.*)', line):
                    val = re.search('.*->(.*)', line)
                    val = val.groups()[0]
                    if isdefined(self.inputs.output_dir):
                        output_dir = self.inputs.output_dir
                    else:
                        output_dir = self._gen_filename('output_dir')
                    val = os.path.join(output_dir, val)
                    if os.path.exists(val):
                        out_file = val

                if out_file:
                    if out_file not in files:
                        files.append(out_file)
                        last_added_file = out_file
                    continue

                if line.startswith("Reorienting as "):
                    reoriented_files.append(line[len("Reorienting as "):])
                    skip = True
                    continue
                elif line.startswith("Cropping NIfTI/Analyze image "):
                    base, filename = os.path.split(
                        line[len("Cropping NIfTI/Analyze image "):])
                    filename = "c" + filename
                    if os.path.exists(os.path.join(
                            base, filename)) or self.inputs.reorient_and_crop:
                        # if reorient&crop is true but the file doesn't exist, this errors when setting outputs
                        reoriented_and_cropped_files.append(
                            os.path.join(base, filename))
                        skip = True
                        continue

            skip = False
        return files, reoriented_files, reoriented_and_cropped_files, bvecs, bvals

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['converted_files'] = self.output_files
        outputs['reoriented_files'] = self.reoriented_files
        outputs[
            'reoriented_and_cropped_files'] = self.reoriented_and_cropped_files
        outputs['bvecs'] = self.bvecs
        outputs['bvals'] = self.bvals
        return outputs

    def _gen_filename(self, name):
        if name == 'output_dir':
            return os.getcwd()
        elif name == 'config_file':
            self._config_created = True
            config_file = "config.ini"
            with open(config_file, "w") as f:
                # disable interactive mode
                f.write("[BOOL]\nManualNIfTIConv=0\n")
            return config_file
        return None


class Dcm2niixInputSpec(CommandLineInputSpec):
    source_names = InputMultiPath(
        File(exists=True),
        argstr="%s",
        position=-1,
        copyfile=False,
        mandatory=True,
        desc=('A set of filenames to be converted. Note that the current ' 
              'version (1.0.20180328) of dcm2niix converts any files in the '
              'directory. To only convert specific files they should be in an '
              'isolated directory'),
        xor=['source_dir'])
    source_dir = Directory(
        exists=True,
        argstr="%s",
        position=-1,
        mandatory=True,
        desc='A directory containing dicom files to be converted',
        xor=['source_names'])
    out_filename = traits.Str(
        argstr="-f %s",
        desc="Output filename template ("
             "%a=antenna (coil) number, "
             "%c=comments, "
             "%d=description, "
             "%e=echo number, "
             "%f=folder name, "
             "%i=ID of patient, "
             "%j=seriesInstanceUID, "
             "%k=studyInstanceUID, "
             "%m=manufacturer, "
             "%n=name of patient, "
             "%p=protocol, "
             "%s=series number, "
             "%t=time, "
             "%u=acquisition number, "
             "%v=vendor, "
             "%x=study ID; "
             "%z=sequence name)")
    output_dir = Directory(
        ".",
        usedefault=True,
        exists=True,
        argstr='-o %s',
        desc="Output directory")
    bids_format = traits.Bool(
        True,
        argstr='-b',
        usedefault=True,
        desc="Create a BIDS sidecar file")
    anon_bids = traits.Bool(
        argstr='-ba',
        requires=["bids_format"],
        desc="Anonymize BIDS")
    compress = traits.Enum(
        'y', 'i', 'n', '3',
        argstr='-z %s',
        usedefault=True,
        desc="Gzip compress images - [y=pigz, i=internal, n=no, 3=no,3D]")
    merge_imgs = traits.Bool(
        False,
        argstr='-m',
        usedefault=True,
        desc="merge 2D slices from same series")
    single_file = traits.Bool(
        False,
        argstr='-s',
        usedefault=True,
        desc="Single file mode")
    verbose = traits.Bool(
        False,
        argstr='-v',
        usedefault=True,
        desc="Verbose output")
    crop = traits.Bool(
        False,
        argstr='-x',
        usedefault=True,
        desc="Crop 3D T1 acquisitions")
    has_private = traits.Bool(
        False,
        argstr='-t',
        usedefault=True,
        desc="Flag if text notes include private patient details")
    compression = traits.Enum(
        1, 2, 3, 4, 5, 6, 7, 8, 9,
        argstr='-%d',
        desc="Gz compression level (1=fastest, 9=smallest)")
    comment = traits.Str(
        argstr='-c %s',
        desc="Comment stored as NIfTI aux_file")
    ignore_deriv = traits.Bool(
        argstr='-i',
        desc="Ignore derived, localizer and 2D images")
    series_numbers = InputMultiPath(
        traits.Str(),
        argstr='-n %s...',
        desc="Selectively convert by series number - can be used up to 16 times")
    philips_float = traits.Bool(
        argstr='-p',
        desc="Philips precise float (not display) scaling")


class Dcm2niixOutputSpec(TraitedSpec):
    converted_files = OutputMultiPath(File(exists=True))
    bvecs = OutputMultiPath(File(exists=True))
    bvals = OutputMultiPath(File(exists=True))
    bids = OutputMultiPath(File(exists=True))


class Dcm2niix(CommandLine):
    """Uses Chris Rorden's dcm2niix to convert dicom files

    Examples
    ========

    >>> from nipype.interfaces.dcm2nii import Dcm2niix
    >>> converter = Dcm2niix()
    >>> converter.inputs.source_dir = 'dicomdir'
    >>> converter.inputs.compression = 5
    >>> converter.inputs.output_dir = 'ds005'
    >>> converter.cmdline
    'dcm2niix -b y -z y -5 -x n -t n -m n -o ds005 -s n -v n dicomdir'
    >>> converter.run() # doctest: +SKIP

    # In the example below, we note that the current version of dcm2niix
    # converts any files in the directory containing the files in the list. We
    # also do not support nested filenames with this option. Thus all files
    # should have a common root directory.
    >>> converter = Dcm2niix()
    >>> converter.inputs.source_names = ['functional_1.dcm', 'functional_2.dcm']
    >>> converter.inputs.compression = 5
    >>> converter.inputs.output_dir = 'ds005'
    >>> converter.cmdline
    'dcm2niix -b y -z y -5 -x n -t n -m n -o ds005 -s n -v n .'
    >>> converter.run() # doctest: +SKIP
    """

    input_spec = Dcm2niixInputSpec
    output_spec = Dcm2niixOutputSpec
    _cmd = 'dcm2niix'

    @property
    def version(self):
        return Info.version()

    def _format_arg(self, opt, spec, val):
        bools = ['bids_format', 'merge_imgs', 'single_file', 'verbose', 'crop',
                 'has_private', 'anon_bids', 'ignore_deriv', 'philips_float']
        if opt in bools:
            spec = deepcopy(spec)
            if val:
                spec.argstr += ' y'
            else:
                spec.argstr += ' n'
                val = True
        if opt == 'source_names':
            return spec.argstr % (os.path.dirname(val[0]) or '.')
        return super(Dcm2niix, self)._format_arg(opt, spec, val)

    def _run_interface(self, runtime):
        # may use return code 1 despite conversion
        runtime = super(Dcm2niix, self)._run_interface(
            runtime, correct_return_codes=(0, 1, ))
        if self.inputs.bids_format:
            (self.output_files, self.bvecs, self.bvals,
             self.bids) = self._parse_stdout(runtime.stdout)
        else:
            (self.output_files, self.bvecs, self.bvals) = self._parse_stdout(
                runtime.stdout)
        return runtime

    def _parse_stdout(self, stdout):
        files = []
        bvecs = []
        bvals = []
        bids = []
        skip = False
        find_b = False
        for line in stdout.split("\n"):
            if not skip:
                out_file = None
                if line.startswith("Convert "):  # output
                    fname = str(re.search('\S+/\S+', line).group(0))
                    out_file = os.path.abspath(fname)
                    # extract bvals
                    if find_b:
                        bvecs.append(out_file + ".bvec")
                        bvals.append(out_file + ".bval")
                        find_b = False
                # next scan will have bvals/bvecs
                elif 'DTI gradients' in line or 'DTI gradient directions' in line or 'DTI vectors' in line:
                    find_b = True
                if out_file:
                    ext = '.nii' if self.inputs.compress == 'n' else '.nii.gz'
                    files.append(out_file + ext)
                    if self.inputs.bids_format:
                        bids.append(out_file + ".json")
            skip = False
        # just return what was done
        if not bids:
            return files, bvecs, bvals
        else:
            return files, bvecs, bvals, bids

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['converted_files'] = self.output_files
        outputs['bvecs'] = self.bvecs
        outputs['bvals'] = self.bvals
        if self.inputs.bids_format:
            outputs['bids'] = self.bids
        return outputs
